package hk.advanpro.android.sdk.device;

import android.support.annotation.NonNull;

import java.io.File;

import hk.advanpro.android.sdk.device.callback.DeviceCommandCallback;
import hk.advanpro.android.sdk.device.callback.DeviceConnectCallback;
import hk.advanpro.android.sdk.device.callback.DeviceDisconnectCallback;
import hk.advanpro.android.sdk.device.callback.DeviceEventCallback;
import hk.advanpro.android.sdk.device.callback.DeviceSettingCallback;
import hk.advanpro.android.sdk.device.callback.DeviceUpgradeCallback;
import hk.advanpro.android.sdk.device.enumeration.DeviceType;
import hk.advanpro.android.sdk.device.params.DeviceCommandParams;
import hk.advanpro.android.sdk.device.params.DeviceSettingParams;
import hk.advanpro.android.sdk.device.result.DeviceScanResult;
import hk.advanpro.android.sdk.device.result.DeviceSettingResult;

/**
 * 安润普设备控制接口。
 * Created by yejianfei on 2017/6/7.
 */

public interface Device {

    /**
     * 已经连接中，不能重复连接，错误码
     */
    int ERROR_ALREADY_CONNECTED = 500;

    /**
     * 设备断开连接，请求命令，错误码
     */
    int ERROR_IS_DISCONNECTED = 501;

    /**
     * 服务不存在
     */
    int ERROR_SERVICE_NOT_EXISTS = 502;

    /**
     * 特征不存在
     */
    int ERROR_CHARACTERISTIC_NOT_EXISTS = 503;
    /**
     * 不支持改功能
     */
    int ERROR_FUNCTION_NOT_SUPPORT = 504;
    /**
     * 发现服务超时
     */
    int ERROR_SERVICE_DISCOVERY_TIMEOUT = 505;

    /**
     * 上一个命令还没请求完成，不能请求相同命令，错误码
     */
    int ERROR_LAST_COMMAND_NOT_DONE = 601;

    /**
     * 请求超时错误码
     */
    int ERROR_IS_TIME_OUT = 701;

    /**
     * 已经进行升级操作，不能同事升级，错误码
     */
    int ERROR_ALREADY_UPGRADE = 801;

    /**
     * 文件不存在
     */
    int ERROR_UPGRADE_FILE_NOT_EXISTS = 810;
    /**
     * 升级其他错误
     */
    int ERROR_UPGRADE_OTHER_ERROR = 899;

    /**
     * 传入设备扫描信息，打开设备，该操作不会真正连接设备，用于设备连接前的初始化操作。
     *
     * @param result 设备扫描信息
     */
    void open(@NonNull DeviceScanResult result);

    /**
     * 设备连接操作，用于连接真正的物理设备。
     *
     * @param cb 设备连接回调
     */
    void connect(@NonNull DeviceConnectCallback cb);

    /**
     * 设备连接操作，自定义超时
     *
     * @param timeOut 超时，单位秒
     * @param cb      设备连接回调
     */
    void connect(int timeOut, @NonNull DeviceConnectCallback cb);


    /**
     * 设备连接操作，用于连接真正的物理设备。
     *
     * @param result 设备扫描信息
     * @param cb     设备连接回调
     */
    void connect(@NonNull DeviceScanResult result, @NonNull DeviceConnectCallback cb);


    /**
     * 设备断开连接操作，用于断开真的物理设备。
     *
     * @param cb 设备断开连接回调
     */
    void disconnect(@NonNull DeviceDisconnectCallback cb);

    /**
     * 用于监听设备事件或消息推送的方法。
     *
     * @param key 事件或消息名称
     * @param cb  推送回调
     */
    void on(@NonNull String key, @NonNull DeviceEventCallback cb);

    /**
     * 用于取消设备事件监听的方法。
     *
     * @param key 事件或消息名称
     * @param cb  待取消的事件回调方法
     */
    void un(@NonNull String key, @NonNull DeviceEventCallback cb);

    /**
     * 用于取消设备全部事件监听的方法。
     */
    void unAllEvent();

    /**
     * 用于发送设备命令的操作。
     *
     * @param key    命令字
     * @param params 命令参数
     * @param cb     命令响应回调接口
     */
    void command(@NonNull String key, @NonNull DeviceCommandParams params, @NonNull DeviceCommandCallback cb);

    /**
     * 用于发送设备命令的操作。
     *
     * @param key     命令字
     * @param params  命令参数
     * @param timeOut 超时时间（秒） 小于等于0 不设定超时时间，大于0 设定超时时间
     * @param cb      命令响应回调接口
     */
    void command(@NonNull String key, @NonNull DeviceCommandParams params, int timeOut, @NonNull DeviceCommandCallback cb);

    /**
     * 用于发送设备启动命令的操作。
     *
     * @param params 命令参数
     * @param cb     命令响应回调接口
     */
    void start(@NonNull DeviceCommandParams params, @NonNull DeviceCommandCallback cb);

    /**
     * 用于发送设备停止命令的操作。
     *
     * @param params 命令参数
     * @param cb     命令响应回调接口
     */
    void stop(DeviceCommandParams params, @NonNull DeviceCommandCallback cb);

    /**
     * 用于发送设备停止命令的操作。
     * 参数默认为 null
     *
     * @param cb 命令响应回调接口
     */
    void stop(@NonNull DeviceCommandCallback cb);

    /**
     * 用于发送设备配置命令的操作。
     *
     * @param params 命令参数
     * @param cb     命令响应回调接口
     */
    void setting(@NonNull DeviceSettingParams params, @NonNull DeviceSettingCallback cb);

    /**
     * 用于获取设备配置命令的操作。
     *
     * @param cb 命令响应回调接口
     */
    void readSetting(@NonNull DeviceSettingCallback cb);

    /**
     * 获取设备配置，从缓存拿
     *
     * @return 设备配置
     */
    DeviceSettingResult getSetting();

    /**
     * 用于升级固件的操作
     *
     * @param file 升级固件文件
     * @param cb   命令响应回调接口
     * @return 是否开启升级
     */
    boolean upgrade(@NonNull File file, @NonNull DeviceUpgradeCallback cb);

    /**
     * 设备是否已经连接
     *
     * @return 是否连接
     */
    boolean isConnected();

    /**
     * 设备是否已经关闭，
     *
     * @return boolean
     */
    boolean isClose();

    /**
     * 设备信息是否已读取
     *
     * @return 是否读取
     */
    boolean isInfoRead();

    /**
     * 设备是否正在升级固件
     *
     * @return 是否升级中
     */
    boolean isUpgrading();

    /**
     * 设置设备自动断开连接后是否自动重连
     *
     * @param enable true：重连 ，false:不重连
     */
    void setEnableReConnect(boolean enable);

    /**
     * 通过设备信息常量，获取设备信息。
     *
     * @param key 设备信息常量
     * @return 设备信息
     */
    String getInfo(@NonNull String key);

    /**
     * 获取设备地址
     *
     * @return 设备地址
     */
    String getAddress();

    /**
     * 设备类型
     *
     * @return 设备类型
     */
    DeviceType getDeviceType();

    /**
     * 读取设备信息
     */
    void readDeviceInfo();


    /**
     * 是否为应用模式
     */
    boolean isApplicationMode();

    /**
     * 是否开启采集
     *
     * @return true 开启采集
     */
    boolean isStarted();

    /**
     * 是否连接时候获取采集状态（默认不获取）
     *
     * @return 是否连接时候获取采集状态
     */
    boolean isReadStateWhenConnected();


    void setReadStateWhenConnected(boolean readStateWhenConnected);

}
