package hk.advanpro.android.sdk.device.ble;

import android.bluetooth.BluetoothAdapter;
import android.bluetooth.BluetoothDevice;
import android.bluetooth.BluetoothManager;
import android.content.Context;
import android.support.annotation.NonNull;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;
import java.util.Timer;
import java.util.TimerTask;

import hk.advanpro.android.core.log.ALogger;
import hk.advanpro.android.sdk.commons.ALog;
import hk.advanpro.android.sdk.device.Device;
import hk.advanpro.android.sdk.device.DeviceManager;
import hk.advanpro.android.sdk.device.callback.BatchDeviceConnectCallback;
import hk.advanpro.android.sdk.device.callback.DeviceCallbackException;
import hk.advanpro.android.sdk.device.callback.DeviceConnectCallback;
import hk.advanpro.android.sdk.device.callback.DeviceManagerScanCallback;
import hk.advanpro.android.sdk.device.enumeration.ConnectType;
import hk.advanpro.android.sdk.device.enumeration.DeviceMode;
import hk.advanpro.android.sdk.device.enumeration.DeviceType;
import hk.advanpro.android.sdk.device.result.DeviceConnectResult;
import hk.advanpro.android.sdk.device.result.ble.BLEBatchDeviceConnectResult;
import hk.advanpro.android.sdk.device.result.ble.BLEDeviceScanResult;
import hk.advanpro.android.sdk.utils.CommonUtils;

/**
 * 低耗功蓝牙设备管理类
 * Created by yejianfei on 2017/6/7.
 */
@SuppressWarnings("deprecation")
public class BLEDeviceManager implements DeviceManager<BLEDeviceScanResult> {

    private static final String TAG = BLEDeviceManager.class.getSimpleName();

    private Context context;

    private BluetoothManager bluetoothManager;

    private BluetoothAdapter bluetoothAdapter;

    private Set<BLEDeviceScanResult.ScanRecord> records = new LinkedHashSet<>();

    private DeviceManagerScanCallback<BLEDeviceScanResult> scanCallback;
    private DeviceType deviceType;

    private boolean scanning = false;
    private Timer timer;


    private BluetoothAdapter.LeScanCallback callback = new BluetoothAdapter.LeScanCallback() {
        @Override
        public void onLeScan(BluetoothDevice device, int rssi, byte[] scanRecord) {
            ALog.d(TAG, String.format("ls scan device name %s [%s] and rssi %d scanRecord %s", device.getName(), device.getAddress(), rssi, CommonUtils.bytesToHexString(scanRecord)));
            BLEDeviceScanResult.ScanRecord record = createScanRecord(device, rssi, scanRecord);
            if (scanning && record != null && !records.contains(record) && (deviceType == null || deviceType.equals(record.getManufacturer().getType()))) {
                records.add(record);
                BLEDeviceScanResult result = new BLEDeviceScanResult(context, record, device);
                scanCallback.onScanning(result);

            }

        }
    };

    public Context getContext() {
        return context;
    }

    public BluetoothManager getBluetoothManager() {
        return bluetoothManager;
    }

    public BluetoothAdapter getBluetoothAdapter() {
        return bluetoothAdapter;
    }

    public synchronized void init(Context context) {
        this.context = context;
        bluetoothManager = (BluetoothManager) context.getSystemService(Context.BLUETOOTH_SERVICE);
        bluetoothAdapter = bluetoothManager.getAdapter();
    }

    /**
     * 进行设备扫描，蓝牙使用系统扫描机制，串口获取系统串口列表。
     *
     * @param timeout 超时时间 单位秒
     * @param cb      回调接口
     */
    @Override
    public synchronized void scan(final int timeout, final DeviceManagerScanCallback<BLEDeviceScanResult> cb) {
        this.scan(timeout, null, cb);
    }

    @Override
    public void scan(int timeout, DeviceType filterType, DeviceManagerScanCallback<BLEDeviceScanResult> cb) {
        this.scanCallback = cb;
        this.deviceType = filterType;
        //日志埋点，搜索设备，0：开始
        ALogger.i("scan", "t", "ble", "st", "0");

        if (timeout > 0) {
            timer = new Timer();
            timer.schedule(new TimerTask() {
                @Override
                public void run() {
                    if (scanning) {
                        stop();
                    }

                }
            }, timeout * 1000);
        }
        scanning = true;
        bluetoothAdapter.startLeScan(callback);
    }

    @Override
    public synchronized void stop() {
        //日志埋点，搜索设备，2：结束
        ALogger.i("scan", "t", "ble", "st", "2");
        scanning = false;
        bluetoothAdapter.stopLeScan(callback);
        if (timer != null) {
            timer.cancel();
            timer = null;
        }
        if (scanCallback != null) {
            scanCallback.onStop();
            scanCallback = null;
        }
        records.clear();
    }

    @Override
    public void instance(final DeviceConnection connection, final DeviceManagerScanCallback<BLEDeviceScanResult> cb) {
        instance(connection, 30, cb);
    }

    @Override
    public void instance(final DeviceConnection connection, final int timeout, final DeviceManagerScanCallback<BLEDeviceScanResult> cb) {
        //日志埋点，搜索设备，0：开始
        ALogger.i("scan", "t", "ble", "st", "0");

        final Timer timer = new Timer();
        final BluetoothAdapter.LeScanCallback callback = new BluetoothAdapter.LeScanCallback() {
            boolean isStop = false;

            @Override
            public void onLeScan(BluetoothDevice device, int rssi, byte[] scanRecord) {
                ALog.d(TAG, String.format("instance ls scan device name %s [%s] and rssi %d scanRecord %s", device.getName(), device.getAddress(), rssi, CommonUtils.bytesToHexString(scanRecord)));
                if (!isStop && device.getAddress().equals(connection.getAddress())) {
                    //日志埋点，搜索设备，1：搜到指定设备
                    ALogger.i("scan", "t", "ble", "st", "1", "addr", device.getAddress(), "rssi", String.valueOf(rssi));
                    isStop = true;
                    timer.cancel();
                    bluetoothAdapter.stopLeScan(this);

                    BLEDeviceScanResult.ScanRecord record = createScanRecord(device, rssi, scanRecord);
                    BLEDeviceScanResult result = new BLEDeviceScanResult(context, record, device);
                    cb.onScanning(result);
                    cb.onStop();
                    //日志埋点，搜索设备，2：结束
                    ALogger.i("scan", "t", "ble", "st", "2");
                }
            }
        };
        timer.schedule(new TimerTask() {
            @Override
            public void run() {
                //日志埋点，搜索设备，2：结束
                ALogger.e("scan", "t", "ble", "st", "2", "err", String.valueOf(Device.ERROR_IS_TIME_OUT));
                cb.onError(new DeviceCallbackException(Device.ERROR_IS_TIME_OUT, "timeout=" + timeout + "s"));
                bluetoothAdapter.stopLeScan(callback);
            }
        }, timeout * 1000);
        bluetoothAdapter.startLeScan(callback);
    }

    protected BLEDeviceScanResult.ScanRecord createScanRecord(BluetoothDevice device, int rssi, byte[] record) {
        short flag = 0;
        short uuid = 0;
        String localName = null;
        byte[] manufacturer = null;

        ByteBuffer buffer = ByteBuffer.wrap(record).order(ByteOrder.LITTLE_ENDIAN);
        while (buffer.remaining() > 2) {
            byte len = buffer.get();

            if (len <= 0 || len > buffer.remaining()) {
                break;
            }

            byte type = buffer.get();
            byte[] val = new byte[len - 1];

            switch (type) {
                case 0x01:
                    flag = buffer.get();
                    break;
                case 0x02:
                    uuid = buffer.getShort();
                    break;
                case 0x03:
                    uuid = buffer.getShort();
                    break;
                case 0x09:
                    buffer.get(val);
                    localName = new String(val);
                    break;
                case (byte) 0xff:
                    buffer.get(val);
                    manufacturer = val;
                    break;
                default:
                    buffer.get(val);
                    break;

            }
        }

        if ((manufacturer == null || (manufacturer[0] != (byte) 0xc1 && manufacturer[0] != (byte) 0xc2))) {
            return null;
        } else {
            BLEDeviceScanResult.ScanRecord info = new BLEDeviceScanResult.ScanRecord();
            info.setUuid(uuid);
            info.setLocalName(localName);
            info.setRssi(rssi);
            info.setAddress(device.getAddress());

            if (manufacturer != null) {
                BLEDeviceScanResult.ScanRecordManufacturer data = new BLEDeviceScanResult.ScanRecordManufacturer();
                data.setCompany(String.format("%02x", (short) (manufacturer[0] & 0xff)).toUpperCase());
                int type = (manufacturer[1] & 0xff);
                DeviceType deviceType = DeviceType.Unknown;
                switch (type) {
                    case 1:
                        deviceType = DeviceType.Belt;// 呼吸带
                        break;
                    case 2:
                        deviceType = DeviceType.Band;// 腰带
                        break;
                    case 3:
                        deviceType = DeviceType.SMan;// 内裤
                        break;
                    case 4:
                        deviceType = DeviceType.Insole; // 鞋垫
                        break;
                    case 7:
                        deviceType = DeviceType.Futon;// 跪垫
                        break;
                    case 8:
                        deviceType = DeviceType.Kedian; // 磕垫
                        break;
                    case 10:
                        deviceType = DeviceType.SDP; // 开发版
                        break;
                    case 11:
                        deviceType = DeviceType.SDP_Insole; // 开发版鞋垫
                        break;
                    case 13:
                        deviceType = DeviceType.HeartShirt; // 心跳衣
                        break;
                    case 14:
                        deviceType = DeviceType.BreathShirt; // 呼吸衣
                        break;
                }
                data.setType(deviceType);
                String address = String.format("%02x", manufacturer[3]) + ":" +
                        String.format("%02x", manufacturer[4]) + ":" +
                        String.format("%02x", manufacturer[5]) + ":" +
                        String.format("%02x", manufacturer[6]) + ":" +
                        String.format("%02x", manufacturer[7]) + ":" +
                        String.format("%02x", manufacturer[8]);
                data.setAddress(address.toUpperCase());

                byte product = (byte) (manufacturer[2] & 0xff);
                data.setMode((product >> 7) == 0 ? DeviceMode.Application : DeviceMode.BootLoader);
                data.setVersion((short) ((product >> 3) & 0x7));
                data.setMajor((short) (product & 0x7));

                HashMap<String, String> map = new HashMap<>();
                switch (data.getType()) {
                    case Insole:
                    case SDP_Insole:
                        map.put("foot", String.valueOf((product >> 6) & 0B1));//0左脚 1右脚
                        break;
                }

                data.setSerial(String.format("%02x%02x%02x", manufacturer[0], manufacturer[1], manufacturer[2]).toUpperCase());
                data.setDescription(map);
                info.setManufacturer(data);
            }

            return info;
        }
    }

    @Override
    public boolean isEnable(ConnectType type) {
        switch (type) {
            case BLE:
                return bluetoothManager != null && bluetoothAdapter != null && bluetoothAdapter.enable();
        }
        return false;
    }

    @Override
    public void connect(int timeOut, @NonNull BatchDeviceConnectCallback callback, DeviceConnection... connections) {
        final int scanTime = timeOut;
        final int connectTime = timeOut;
        BatchDeviceConnect batchDeviceConnect = new BatchDeviceConnect(this);
        batchDeviceConnect.start(scanTime, connectTime, callback, connections);
    }

    @Override
    public boolean isScanning() {
        return scanning;
    }


    /**
     * 批量设备连接器，内部类
     */
    protected static class BatchDeviceConnect {
        private int scanTime;//扫描时间
        private int connectTime;//连接时间
        private DeviceManager deviceManager;
        private BatchDeviceConnectCallback callback;

        private List<DeviceConnection> connectionList;
        private List<BLEBatchDeviceConnectResult.ConnectResult> connectResultList;
        private List<BLEBatchDeviceConnectResult.ConnectResult> finishList;

        private int totalCount;
        private int createdCount;
        private int connectCount;


        public BatchDeviceConnect(DeviceManager deviceManager) {
            this.deviceManager = deviceManager;
        }

        protected void init() {
            connectionList = new ArrayList<>();
            connectResultList = new ArrayList<>();
            finishList = new ArrayList<>();
            totalCount = 0;
            createdCount = 0;
            connectCount = 0;
        }

        protected void start(int scanTime, int connectTime, final BatchDeviceConnectCallback callback, DeviceConnection... connections) {
            this.scanTime = scanTime;
            this.connectTime = connectTime;
            this.callback = callback;
            init();

            //先遍历待连接的设备
            for (DeviceConnection connection : connections) {
                if (connectionList.contains(connection)) {
                    //过滤重复的设备信息
                    continue;
                }
                //添加到待连接设备列表
                connectionList.add(connection);
                BLEBatchDeviceConnectResult.ConnectResult result = new BLEBatchDeviceConnectResult.ConnectResult();
                result.setConnection(connection);
                connectResultList.add(result);
                //总待连接设备个数累计
                totalCount++;
            }
            //如果待连接设备个数为0，直接结束
            if (this.totalCount == 0) {
                onFinish();
                return;
            }
            //去搜索设备
            gotoScanDevice();

        }

        /**
         * 去搜索设备
         */
        protected void gotoScanDevice() {
            //开始搜索设备
            deviceManager.scan(this.scanTime, new DeviceManagerScanCallback<BLEDeviceScanResult>() {
                @Override
                public void onScanning(BLEDeviceScanResult result) {
                    String address = result.getAddress();
                    if (address == null) {
                        return;
                    }
                    //搜索一个设备，遍历并查找是否在待连接设备列表
                    for (BLEBatchDeviceConnectResult.ConnectResult r : connectResultList) {
                        //在待连接设备列表
                        if (address.equalsIgnoreCase(r.getConnection().getAddress())) {
                            //创建设备
                            r.setDevice(result.create());
                            //创建设备个数累加
                            createdCount++;
                            if (callback != null)
                                callback.onCreated(r.getConnection(), r.getDevice());
                            break;
                        }
                    }

                    //检查是否需要停止搜索设备
                    if (checkFinishScan()) {
                        deviceManager.stop();
                    }
                }

                @Override
                public void onStop() {
                    //如果创建设备个数为0，直接结束搜索
                    if (createdCount == 0) {
                        onFinish();
                        return;
                    }
                    //遍历需要连接设备，调用去连接方法
                    for (BLEBatchDeviceConnectResult.ConnectResult result : connectResultList) {
                        if (!result.isCreated()) {
                            //往已完成队列添加设备
                            finishList.add(result);
                            continue;
                        }
                        gotoConnectDevice(result.getDevice());
                    }
                }

                @Override
                public void onError(DeviceCallbackException error) {
                    if (callback != null)
                        callback.onError(error);
                }
            });
        }

        protected boolean checkFinishScan() {
            //如果待搜索设备全部搜索完成，停止搜索
            return createdCount == totalCount;
        }

        /**
         * 去连接设备
         *
         * @param device 设备
         */
        protected void gotoConnectDevice(Device device) {
            device.connect(connectTime, new DeviceConnectCallback() {
                @Override
                public void onSucceed(Device device, DeviceConnectResult result) {
                    //设备连接成功
                    synchronized (finishList) {
                        for (BLEBatchDeviceConnectResult.ConnectResult c : connectResultList) {
                            if (c.getDevice() == device && !c.isConnect()) {
                                //找到设备结果，修改设备信息
                                c.setConnect(true);
                                //往已完成队列添加设备
                                finishList.add(c);
                                connectCount++;
                                if (callback != null)
                                    callback.onConnectSuccess(c.getConnection(), device);
                                break;
                            }
                        }
                        checkFinish();
                    }
                }

                @Override
                public void onError(Device device, DeviceCallbackException error) {
                    //连接设备
                    synchronized (finishList) {
                        for (BLEBatchDeviceConnectResult.ConnectResult c : connectResultList) {
                            if (c.getDevice() == device) {
                                c.setException(error);
                                //往已完成队列添加设备
                                finishList.add(c);
                                break;
                            }
                        }
                        checkFinish();
                    }
                }
            });
        }

        /**
         * 检查是否需要结束
         */
        protected void checkFinish() {
            //完成个数，等于总个数
            if (finishList.size() == totalCount) {
                //调用结束方法
                onFinish();
            }
        }

        /**
         * 结束连接设备,并回调
         */
        protected void onFinish() {
            BLEBatchDeviceConnectResult result = new BLEBatchDeviceConnectResult();
            result.setTotalCount(totalCount);
            result.setCreatedCount(createdCount);
            result.setConnectCount(connectCount);
            result.setConnectResultList(connectResultList);
            if (callback != null)
                callback.onFinish(result);
        }
    }
}
