package hk.advanpro.android.sdk.device.ble;

import android.bluetooth.BluetoothGatt;
import android.bluetooth.BluetoothGattCallback;
import android.bluetooth.BluetoothGattCharacteristic;

import java.io.File;
import java.util.UUID;

import hk.advanpro.android.sdk.AdvanproAndroidSDK;
import hk.advanpro.android.sdk.device.Device;
import hk.advanpro.android.sdk.device.callback.DeviceCallbackException;
import hk.advanpro.android.sdk.device.callback.DeviceCommandCallback;
import hk.advanpro.android.sdk.device.callback.DeviceUpgradeCallback;
import hk.advanpro.android.sdk.device.ota.BLEFirmwareUpgradeBase;
import hk.advanpro.android.sdk.device.ota.DeviceFirmwareUpgradeCallback;
import hk.advanpro.android.sdk.device.ota.dfu.BLEDfuFirmwareUpgrade;
import hk.advanpro.android.sdk.device.params.DeviceCommandParams;
import hk.advanpro.android.sdk.device.protocol.BLEShirtProtocol;
import hk.advanpro.android.sdk.device.result.DeviceSettingResult;

/**
 * 智能衣蓝牙设备
 * 呼吸衣、心电衣
 * 作者：jalin on 2017/12/25 10:15
 * 邮箱：zhouxiaolin@advanpro.com
 */
public class BLEShirtDevice extends DefaultBLEDevice {

    /**
     * 心率实时数据改变事件
     */
    public static final String EVENT_SHIRT_HEART_RATE = "EVENT_SHIRT_HEART_RATE";

    /**
     * 织物传感器实时数据改变事件
     */
    public static final String EVENT_SHIRT_REAL_SENSOR = "EVENT_SHIRT_REAL_SENSOR";

    //--------------- 蓝牙服务通道 --------------
    /**
     * 自定义服务
     */
    protected static final UUID UUID_SERVICE_PRIMARY = UUID.fromString("00008001-0000-1000-8000-00805F9B34FB");

    /**
     * 呼吸衣，织物传感器实时数据通道 ，Notify
     */
    protected static final UUID UUID_CHARACTERISTIC_REAL_SENSOR = UUID.fromString("00008002-0000-1000-8000-00805F9B34FB");
    /**
     * 电阻值, Read
     */
    protected static final UUID UUID_RANGE_CHARACTERISTIC_CHARACTERISTIC = UUID.fromString("00008006-0000-1000-8000-00805F9B34FB");
    /**
     * 用于改变设备蓝牙广播名称，需要开启写入权限
     */
    protected static final UUID UUID_CHARACTERISTIC_PIPE = UUID.fromString("00008007-0000-1000-8000-00805F9B34FB");
    /**
     * 心电衣，心电值服务，Heart Rate Service
     */
    protected static final UUID UUID_SERVICE_HEART_RATE = UUID.fromString("0000180D-0000-1000-8000-00805F9B34FB");
    /**
     * 心电衣，心电值通道，Heart Rate characteristic , Notify
     */
    protected static final UUID UUID_CHARACTERISTIC_HEART = UUID.fromString("00008018-0000-1000-8000-00805F9B34FB");


    /**
     * 柔性传电阻量程数据,此数据在设备运行过程中不会更改
     */
    private int fssRange = 0;

    private final DefaultBluetoothGattCallback callback = new DefaultBluetoothGattCallback() {
        @Override
        public void onServicesDiscovered(BluetoothGatt gatt, int status) {
            super.onServicesDiscovered(gatt, status);
            if (isApplicationMode()) {
                // 应用模式

                // 读取织物传感器电阻值
                read(UUID_SERVICE_PRIMARY, UUID_RANGE_CHARACTERISTIC_CHARACTERISTIC);

                // 打开监听织物传感器实时数据通道
                BLEShirtDevice.this.notify(UUID_SERVICE_PRIMARY, UUID_CHARACTERISTIC_REAL_SENSOR, true);

                // 打开监听心率通道
                BLEShirtDevice.this.notify(UUID_SERVICE_HEART_RATE, UUID_CHARACTERISTIC_HEART, true);
            }
        }

        @Override
        public void onCharacteristicChanged(BluetoothGatt gatt, BluetoothGattCharacteristic characteristic) {
            super.onCharacteristicChanged(gatt, characteristic);

            if (UUID_CHARACTERISTIC_HEART.equals(characteristic.getUuid())) {
                // 心率实时数据
                fireEvent(EVENT_SHIRT_HEART_RATE, BLEShirtProtocol.buildHeartRateEventResult(characteristic.getValue()));
            } else if (UUID_CHARACTERISTIC_REAL_SENSOR.equals(characteristic.getUuid())) {
                // 织物传感器实时数据
                fireEvent(EVENT_SHIRT_REAL_SENSOR, BLEShirtProtocol.buildRealSensorEventResult(fssRange, characteristic.getValue()));
            }

        }

        @Override
        public void onCharacteristicRead(BluetoothGatt gatt, BluetoothGattCharacteristic characteristic, int status) {
            super.onCharacteristicRead(gatt, characteristic, status);

            if (characteristic.getUuid().equals(UUID_RANGE_CHARACTERISTIC_CHARACTERISTIC)) {
                // 读取电阻
                int range = BLEShirtProtocol.buildFssRange(characteristic.getValue());
                fssRange = range;
            }
        }
    };

    @Override
    public DeviceSettingResult getSetting() {
        return null;
    }

    @Override
    protected CommandChannelInfo getCommandChannelInfo(String key) {
        return super.getCommandChannelInfo(key);
    }

    @Override
    protected BluetoothGattCallback getBluetoothGattCallback() {
        return callback;
    }

    @Override
    public void start(DeviceCommandParams params, DeviceCommandCallback cb) {
        throw new UnsupportedOperationException();
    }

    @Override
    public void stop(DeviceCommandParams params, DeviceCommandCallback cb) {
        throw new UnsupportedOperationException();
    }

    @Override
    public boolean upgrade(final File file, final DeviceUpgradeCallback cb) {
        if (this.isUpgrading()) {
            //不能重复升级
            cb.onError(this, new DeviceCallbackException(Device.ERROR_ALREADY_UPGRADE));
            return false;
        }
        if (!file.exists()) {
            //文件不存在
            cb.onError(this, new DeviceCallbackException(Device.ERROR_UPGRADE_FILE_NOT_EXISTS));
            return false;
        }
        if (!this.isConnected()) {
            //升级之前先连接设备
            cb.onError(this, new DeviceCallbackException(Device.ERROR_IS_DISCONNECTED));
            return false;
        }

        DeviceFirmwareUpgradeCallback callback = new DeviceFirmwareUpgradeCallback() {
            @Override
            public void start(Device device) {

            }

            @Override
            public void onStatusChange(Device device, int status) {

            }

            @Override
            public void onProgress(Device device, long total, long progress) {
                cb.onProgress(device, total, progress);
            }

            @Override
            public void onSuccess(Device device) {
                isUpgrading = false;
                cb.onSuccess(device, null);

            }

            @Override
            public void onError(Device device, DeviceCallbackException error) {
                isUpgrading = false;
                cb.onError(device, error);
            }
        };
        isUpgrading = true;
        // DFU 固件升级
        BLEFirmwareUpgradeBase bleFirmwareUpgradeBase = new BLEDfuFirmwareUpgrade();
        bleFirmwareUpgradeBase.start(AdvanproAndroidSDK.getAppContext(), this, file, callback);
        return true;
    }

    private boolean isUpgrading = false;

    @Override
    public boolean isUpgrading() {
        return isUpgrading;
    }

    @Override
    public boolean isStarted() {
        return false;
    }

    @Override
    protected void afterRegisterCommand(String key, DeviceCommandCallback cb) {

    }

    @Override
    protected void onCommandTimeOut(String key) {

    }

}
