package hk.advanpro.android.sdk.device.ble;

import android.bluetooth.BluetoothGatt;
import android.bluetooth.BluetoothGattCallback;
import android.bluetooth.BluetoothGattCharacteristic;

import java.util.HashMap;
import java.util.Map;
import java.util.UUID;

import hk.advanpro.android.core.log.ALogger;
import hk.advanpro.android.sdk.commons.ALog;
import hk.advanpro.android.sdk.device.Device;
import hk.advanpro.android.sdk.device.callback.DeviceCallbackException;
import hk.advanpro.android.sdk.device.callback.DeviceCommandCallback;
import hk.advanpro.android.sdk.device.callback.DeviceConnectCallback;
import hk.advanpro.android.sdk.device.callback.DeviceDisconnectCallback;
import hk.advanpro.android.sdk.device.callback.DeviceSettingCallback;
import hk.advanpro.android.sdk.device.enumeration.DeviceMode;
import hk.advanpro.android.sdk.device.params.DeviceCommandParams;
import hk.advanpro.android.sdk.device.result.DeviceCommandResult;
import hk.advanpro.android.sdk.device.result.DeviceSettingResult;
import hk.advanpro.android.sdk.device.result.ble.BLEDeviceInfoResult;
import hk.advanpro.android.sdk.utils.CommonUtils;

import static hk.advanpro.android.sdk.device.protocol.BLEProtocol.buildEnableWritePermissioneResult;
import static hk.advanpro.android.sdk.device.protocol.BLEProtocol.buildSendPermissionResult;
import static hk.advanpro.android.sdk.device.protocol.BLEProtocol.buildWriteDeviceCodeResult;
import static hk.advanpro.android.sdk.device.protocol.BLEProtocol.buildWriteDeviceIdResult;

/**
 * 低耗功蓝牙设备基类
 * Created by yejianfei on 2017/6/7.
 */
@SuppressWarnings("unchecked")
abstract public class DefaultBLEDevice extends BLEDevice {

    public static final String EVENT_BATTERY_CHANGE = "EVENT_BATTERY_CHANGE";

    public static final String CMD_ENABLE_WRITE_PERMISSION = "CMD_ENABLE_WRITE_PERMISSION";

    public static final String CMD_SEND_PERMISSION_CODE = "CMD_SEND_PERMISSION_CODE";

    public static final String CMD_WRITE_DEVICE_CODE = "CMD_WRITE_DEVICE_CODE";

    public static final String CMD_WRITE_DEVICE_ID = "CMD_WRITE_DEVICE_ID";

    public static final String CMD_GET_INFO = "CMD_GET_INFO";


    /**
     * 蓝牙服务UUID，获取设备信息 , device info service
     */
    protected static final UUID INFO_SERVICE_UUID = UUID.fromString("0000180A-0000-1000-8000-00805F9B34FB");

    /**
     * 蓝牙通道UUID，设备ID,read ID Number
     */
    protected static final UUID ID_INFO_CHARACTERISTIC_UUID = UUID.fromString("00002A24-0000-1000-8000-00805F9B34FB");

    /**
     * 读取设备名称 蓝牙通道UUID,read ,Serial（CODE） Number
     */
    protected static final UUID CODE_CHARACTERISTIC_UUID = UUID.fromString("00002A25-0000-1000-8000-00805F9B34FB");

    /**
     * 读取固件版本 蓝牙通道UUID，read ,Firmware Revision
     */
    protected static final UUID FIRMWARE_CHARACTERISTIC_UUID = UUID.fromString("00002A26-0000-1000-8000-00805F9B34FB");

    /**
     * 读取硬件版本 蓝牙通道UUID，read ,Hardware Revision
     */
    protected static final UUID HARDWARE_CHARACTERISTIC_UUID = UUID.fromString("00002A27-0000-1000-8000-00805F9B34FB");


    /**
     * 蓝牙电量服务,Battery Service
     */
    protected static final UUID BATTERY_SERVICE_UUID = UUID.fromString("0000180F-0000-1000-8000-00805F9B34FB");
    /**
     * 蓝牙电量通道,Battery info
     */
    protected static final UUID BATTERY_CHARACTERISTIC_UUID = UUID.fromString("00002A19-0000-1000-8000-00805F9B34FB");

    /**
     * 自定义蓝牙服务, private service
     */
    protected static final UUID PIPE_SERVICE_UUID = UUID.fromString("00008001-0000-1000-8000-00805F9B34FB");

    /**
     * 蓝牙通道UUID,Notify and Write，用于获取写入权限、写入产品 ID、 VV 等
     */
    protected static final UUID PIPE_CHARACTERISTIC_CHARACTERISTIC = UUID.fromString("0000808F-0000-1000-8000-00805F9B34FB");

    private static final String TAG = BLEDevice.class.getSimpleName();
    private static final Map<Short, String> COMMAND_KEYS = new HashMap<Short, String>() {{
        put((short) 0xffa1, CMD_ENABLE_WRITE_PERMISSION);
        put((short) 0xffb1, CMD_SEND_PERMISSION_CODE);
        put((short) 0x0311, CMD_WRITE_DEVICE_ID);
        put((short) 0x0312, CMD_WRITE_DEVICE_CODE);
    }};
    private BluetoothGattCallback callback = new DefaultBluetoothGattCallback();
    private boolean isInfoRead = false;

    @Override
    public boolean isInfoRead() {
        return isInfoRead;
    }

    @Override
    public synchronized void disconnect(DeviceDisconnectCallback cb) {
        super.disconnect(cb);
    }

    @Override
    public void connect(int timeOut, DeviceConnectCallback cb) {
        //修改设备信息未读取
        isInfoRead = false;
        super.connect(timeOut, cb);
    }

    @Override
    protected CommandChannelInfo getCommandChannelInfo(String key) {
        CommandChannelInfo info = null;

        switch (key) {
            case CMD_ENABLE_WRITE_PERMISSION:
                // 强制使用head + flag模式，原有协议为0xff0101
                info = new CommandChannelInfo(CMD_ENABLE_WRITE_PERMISSION, PIPE_SERVICE_UUID, PIPE_CHARACTERISTIC_CHARACTERISTIC);
                break;
            case CMD_SEND_PERMISSION_CODE:
                // 强制使用head + flag模式，原有协议为0xff0101
                info = new CommandChannelInfo(CMD_SEND_PERMISSION_CODE, PIPE_SERVICE_UUID, PIPE_CHARACTERISTIC_CHARACTERISTIC);
                break;
            case CMD_WRITE_DEVICE_CODE:
                info = new CommandChannelInfo(CMD_WRITE_DEVICE_CODE, PIPE_SERVICE_UUID, PIPE_CHARACTERISTIC_CHARACTERISTIC);
                break;
            case CMD_WRITE_DEVICE_ID:
                info = new CommandChannelInfo(CMD_WRITE_DEVICE_ID, PIPE_SERVICE_UUID, PIPE_CHARACTERISTIC_CHARACTERISTIC);
                break;
            case CMD_GET_INFO:
                info = new CommandChannelInfo(CMD_GET_INFO, INFO_SERVICE_UUID, HARDWARE_CHARACTERISTIC_UUID);
                break;
            default:
        }
        return info;
    }

    @Override
    protected BluetoothGattCallback getBluetoothGattCallback() {
        return callback;
    }

    public class DefaultBluetoothGattCallback extends BLEDeviceBluetoothGattCallback {

        @Override
        public void onServicesDiscovered(BluetoothGatt gatt, int status) {
            super.onServicesDiscovered(gatt, status);

            if (status == BluetoothGatt.GATT_SUCCESS) {
                if (isApplicationMode() && (!isFindService(PIPE_SERVICE_UUID) || !isFindService(INFO_SERVICE_UUID))) {
                    // 应用模式,且服务不存在，直接回调连接失败
                    fireConnected(new DeviceCallbackException(ERROR_SERVICE_NOT_EXISTS));
                    return;
                }
                if (isApplicationMode()) {
                    // 应用模式，notify 通道
                    DefaultBLEDevice.this.notify(PIPE_SERVICE_UUID, PIPE_CHARACTERISTIC_CHARACTERISTIC, true);

                    // 电量通道
                    BluetoothGattCharacteristic bluetoothGattCharacteristic
                            = getBluetoothGattCharacteristic(PIPE_SERVICE_UUID, BATTERY_CHARACTERISTIC_UUID);
                    // 开启电量改变通道
                    if (bluetoothGattCharacteristic == null) {
                        // 兼容旧固件，开启监听电量改变
                        DefaultBLEDevice.this.notify(BATTERY_SERVICE_UUID, BATTERY_CHARACTERISTIC_UUID, true);
                    } else {
                        // 新固件版本，开启监听电量改变
                        DefaultBLEDevice.this.notify(PIPE_SERVICE_UUID, BATTERY_CHARACTERISTIC_UUID, true);
                    }
                }
                //连接成功
                fireConnected(null);
                //读取设备信息
                readDeviceInfo();
            }

        }

        @Override
        public void onCharacteristicChanged(BluetoothGatt gatt, BluetoothGattCharacteristic characteristic) {
            super.onCharacteristicChanged(gatt, characteristic);

            ALog.d(TAG, String.format("[%s] notify %s & %s [%s]", bluetoothDevice.getAddress(), characteristic.getService().getUuid().toString(), characteristic.getUuid().toString(), CommonUtils.bytesToHexString(characteristic.getValue())));
            if (PIPE_CHARACTERISTIC_CHARACTERISTIC.equals(characteristic.getUuid())) {

                byte[] data = characteristic.getValue();
                byte head = data[0];
                byte flag = data[1];
                byte cmd = data[2];

                short commandKey = (short) (((head << 8) & 0xffff) | (flag & 0xff));

                if (head == (byte) 0xff && flag == (byte) 0x01 && cmd == (byte) 0x01) {
                    commandKey = (short) 0xffa1;
                } else if (head == (byte) 0xff && flag == (byte) 0x01 && cmd == (byte) 0x02) {
                    commandKey = (short) 0xffb1;
                }

                switch (commandKey) {
                    case (short) 0xffa1:
                        //提请开启权限命令结果
                        fireCommand(COMMAND_KEYS.get(commandKey), buildEnableWritePermissioneResult(data));
                        break;
                    case (short) 0xffb1:
                        //发送验证数据结果
                        fireCommand(COMMAND_KEYS.get(commandKey), buildSendPermissionResult(data));
                        break;
                    case (short) 0x0311:
                        //写入设备ID结果
                        fireCommand(COMMAND_KEYS.get(commandKey), buildWriteDeviceIdResult(data));
                        break;
                    case (short) 0x0312:
                        //写入用户ID结果
                        fireCommand(COMMAND_KEYS.get(commandKey), buildWriteDeviceCodeResult(data));
                        break;
                    default:
                        break;
                }
            }
            if (BATTERY_CHARACTERISTIC_UUID.equals(characteristic.getUuid())) {
                String oldString = info.get(BLE_DEVICE_INFO_BATTERY);
                String newString = String.valueOf(characteristic.getIntValue(BluetoothGattCharacteristic.FORMAT_SINT8, 0));
                if (!newString.equals(oldString)) {
                    DefaultBLEDevice.this.info.put(BLE_DEVICE_INFO_BATTERY, newString);

                    //日志埋点，电量改变
                    if (isInfoRead)
                        ALogger.i("battery", "id", getInfo(BLEDevice.BLE_DEVICE_INFO_ID), "bat", getInfo(BLE_DEVICE_INFO_BATTERY));

                    fireEvent(EVENT_BATTERY_CHANGE, null);
                }
            }
        }


        @Override
        public void onCharacteristicRead(BluetoothGatt gatt, BluetoothGattCharacteristic characteristic, int status) {
            super.onCharacteristicRead(gatt, characteristic, status);

            if (characteristic.getUuid().equals(ID_INFO_CHARACTERISTIC_UUID)) {
                DefaultBLEDevice.this.info.put(BLE_DEVICE_INFO_ID, new String(characteristic.getValue()));
            }

            if (characteristic.getUuid().equals(CODE_CHARACTERISTIC_UUID)) {
                DefaultBLEDevice.this.info.put(BLE_DEVICE_INFO_CODE, new String(characteristic.getValue()));
            }

            if (characteristic.getUuid().equals(FIRMWARE_CHARACTERISTIC_UUID)) {
                DefaultBLEDevice.this.info.put(BLE_DEVICE_INFO_FIRMWARE, new String(characteristic.getValue()).replace((char) 0x0, ' ').trim());
            }

            if (characteristic.getUuid().equals(HARDWARE_CHARACTERISTIC_UUID)) {
                DefaultBLEDevice.this.info.put(BLE_DEVICE_INFO_HARDWARE, new String(characteristic.getValue()));
            }

            if (characteristic.getUuid().equals(BATTERY_CHARACTERISTIC_UUID)) {
                DefaultBLEDevice.this.info.put(BLE_DEVICE_INFO_BATTERY, String.valueOf(characteristic.getIntValue(BluetoothGattCharacteristic.FORMAT_SINT8, 0)));

                //日志埋点，电量改变
                ALogger.i("battery", "id", getInfo(BLEDevice.BLE_DEVICE_INFO_ID), "bat", getInfo(BLE_DEVICE_INFO_BATTERY));

                fireEvent(EVENT_BATTERY_CHANGE, null);
            }

            if (characteristic.getUuid().equals(BATTERY_CHARACTERISTIC_UUID) || !isApplicationMode()) {
                //设备信息已读取
                DefaultBLEDevice.this.isInfoRead = true;

                //日志埋点，读取设备信息
                ALogger.i("dev_info_read", "addr", getInfo(BLEDevice.BLE_DEVICE_INFO_ADDRESS),
                        "id", getInfo(BLEDevice.BLE_DEVICE_INFO_ID),
                        "code", getInfo(BLE_DEVICE_INFO_CODE),
                        "fw", getInfo(BLE_DEVICE_INFO_FIRMWARE),
                        "hw", getInfo(BLE_DEVICE_INFO_HARDWARE));

                BLEDeviceInfoResult result = new BLEDeviceInfoResult(info);

                //发送设备信息读取成功事件
                fireEvent(EVENT_READ_INFO, result);

                // 回调获取设备信息成功命令
                fireCommand(CMD_GET_INFO, result);
            }
        }

    }

    /**
     * 当前是否正常运行模式
     *
     * @return true=正常运行模式
     */
    @Override
    public boolean isApplicationMode() {
        return (DeviceMode.Application.name().equals(getInfo(BLE_DEVICE_INFO_MODE)));
    }

    @Override
    public synchronized void command(final String key, final DeviceCommandParams params, final int timeOut, final DeviceCommandCallback cb) {
        if (CMD_GET_INFO.equals(key)) {
            // 调用获取设备信息

            BLEDevice.CommandChannelInfo cmd = getCommandChannelInfo(key);
            if (cmd != null && registerCommand(cmd, timeOut, cb)) {
                // 执行命令
                if (isApplicationMode()) {
                    // 应用模式，读取信息
                    // 读取设备ID信息
                    read(INFO_SERVICE_UUID, ID_INFO_CHARACTERISTIC_UUID);
                    // 读取效验信息
                    read(INFO_SERVICE_UUID, CODE_CHARACTERISTIC_UUID);
                    // 读取固件信息
                    read(INFO_SERVICE_UUID, FIRMWARE_CHARACTERISTIC_UUID);
                }

                // ota模式，也可以读取硬件信息
                read(INFO_SERVICE_UUID, HARDWARE_CHARACTERISTIC_UUID);

                if (isApplicationMode()) {
                    // 应用模式 ,读取电池信息
                    BluetoothGattCharacteristic bluetoothGattCharacteristic
                            = getBluetoothGattCharacteristic(PIPE_SERVICE_UUID, BATTERY_CHARACTERISTIC_UUID);

                    if (bluetoothGattCharacteristic == null) {
                        // 兼容旧固件版本，获取蓝牙电量
                        read(BATTERY_SERVICE_UUID, BATTERY_CHARACTERISTIC_UUID);
                    } else {
                        // 新固件版本，获取蓝牙电量
                        read(PIPE_SERVICE_UUID, BATTERY_CHARACTERISTIC_UUID);
                    }
                }

                //命令执行完成
                afterRegisterCommand(cmd.getKey(), cb);
            }
            // 直接 return
            return;
        }
        super.command(key, params, timeOut, cb);
    }

    @Override
    public void readDeviceInfo() {
        //获取设备信息
        command(CMD_GET_INFO, null, new DeviceCommandCallback() {
            @Override
            public void onSuccess(Device device, DeviceCommandResult result) {
                if ((!isApplicationMode() || !isReadStateWhenConnected()) && isNeedConnectedCallback()) {
                    // ota 模式,获取不需要默认获取状态，直接回调连接成功
                    onConnectedCallback(null);
                    return;
                }
                if (isReadStateWhenConnected()) {
                    // 连接成功后获取启动模式
                    // 获取启动状态
                    readSetting(new DeviceSettingCallback() {
                        @Override
                        public void onSuccess(Device device, DeviceSettingResult result) {
                            // 需要连接设备回调
                            if (isNeedConnectedCallback())
                                onConnectedCallback(null);
                        }

                        @Override
                        public void onError(Device device, DeviceCallbackException error) {
                            // 需要连接设备回调
                            if (isNeedConnectedCallback())
                                onConnectedCallback(null);
                        }
                    });
                }

            }

            @Override
            public void onError(Device device, DeviceCallbackException error) {
                if (isNeedConnectedCallback()) {
                    // 连接成功，获取设备信息，直接回调成功
                    onConnectedCallback(null);
                }
            }
        });

    }
}
