package hk.advanpro.android.sdk.device.ota;

import android.content.Context;

import java.io.File;

import hk.advanpro.android.core.log.ALogger;
import hk.advanpro.android.sdk.AdvanproAndroidSDK;
import hk.advanpro.android.sdk.device.Device;
import hk.advanpro.android.sdk.device.DeviceManager;
import hk.advanpro.android.sdk.device.ble.BLEDevice;
import hk.advanpro.android.sdk.device.ble.BLEDeviceManager;
import hk.advanpro.android.sdk.device.callback.DeviceCallbackException;
import hk.advanpro.android.sdk.device.callback.DeviceConnectCallback;
import hk.advanpro.android.sdk.device.callback.DeviceManagerScanCallback;
import hk.advanpro.android.sdk.device.result.DeviceConnectResult;
import hk.advanpro.android.sdk.device.result.ble.BLEDeviceScanResult;

/**
 * 固件升级基类
 * Created by jalin on 2017/12/15.
 */

public abstract class BLEFirmwareUpgradeBase implements DeviceFirmwareUpgrade {

    private Context context;
    private Device device;//设备
    private File file;//升级文件

    private String address;//设备 地址
    private String deviceId;//设备id
    private DeviceFirmwareUpgradeCallback callback;
    private boolean isTransmiting = false;//是否正在传输
    private boolean isTransmitSuccess = false;//是否传输成功

    @Override
    public final void start(Context context, Device device, File file, DeviceFirmwareUpgradeCallback cb) {
        this.context = context;
        this.device = device;
        this.file = file;
        this.callback = cb;
        this.address = device.getInfo(BLEDevice.BLE_DEVICE_INFO_ADDRESS);
        this.deviceId = device.getInfo(BLEDevice.BLE_DEVICE_INFO_ID);
        this.isTransmiting = false;
        this.isTransmitSuccess = false;
        //日志埋点，固件升级 ，0:开始
        ALogger.i("fw_upgrade", "id", deviceId, "mac", address, "s", "0");
        //调用准备升级状态改变
        onStatusChange(STATUS_PREPARE_UPGRADE);
        prepareUpgrade();
    }

    @Override
    public void onStatusChange(int status) {
        switch (status) {
            case STATUS_INTO_BOOTLOADER:
                //日志埋点，固件升级 ，1:进入OTA
                ALogger.i("fw_upgrade", "id", getDeviceId(), "mac", address, "s", "1");
                break;
            case STATUS_OUT_BOOTLOADER:
                //日志埋点，固件升级 ，4:切回运行模式
                ALogger.i("fw_upgrade", "id", getDeviceId(), "mac", address, "s", "4");
                break;
            case STATUS_RECONNECTED:
                if (isTransmitSuccess) {
                    //日志埋点，固件升级 ，5:传输成功后重连上
                    ALogger.i("fw_upgrade", "id", getDeviceId(), "mac", address, "s", "5");
                } else {
                    //日志埋点，固件升级，2:传输前重连上
                    ALogger.i("fw_upgrade", "id", getDeviceId(), "mac", address, "s", "2");
                }
                break;
            case STATUS_TRANSMIT_START:
                //日志埋点，固件升级 ，3:开始传输
                ALogger.i("fw_upgrade", "id", getDeviceId(), "mac", address, "s", "3");
                break;
        }
        callback.onStatusChange(device, status);
    }

    @Override
    public void onTransmitProgress(long total, long progress) {
        if (!isTransmiting && !isTransmitSuccess) {
            //传输开始
            isTransmiting = true;
            onStatusChange(STATUS_TRANSMIT_START);
        }
        //传输进度改变
        callback.onProgress(device, total, progress);
    }

    @Override
    public void onTransmitComplete() {
        isTransmiting = false;
        isTransmitSuccess = true;
        onStatusChange(STATUS_TRANSMIT_COMPLETE);
        //传输完成
        //升级成功后
        // 调用搜索设备方法
        deviceScan();
    }

    @Override
    public void onUpgradeSuccess() {
        //日志埋点，6:固件升级 完成
        ALogger.i("fw_upgrade", "id", getDeviceId(), "mac", address, "s", "6");
        callback.onSuccess(device);
    }

    @Override
    public void onUpgradeError(DeviceCallbackException error) {
        //日志埋点，6:固件升级 完成,报错
        ALogger.e("fw_upgrade", "id", getDeviceId(), "mac", address, "s", "6", "err", String.valueOf(error.getCode()));
        callback.onError(device, error);
    }


    protected void deviceScan() {
        onStatusChange(STATUS_BLE_SCAN);
        //重新扫描设备
        BLEDeviceManager manager = AdvanproAndroidSDK.getDeviceManager(BLEDeviceManager.class);
        manager.instance(new DeviceManager.DeviceConnection(address), 30, new DeviceManagerScanCallback<BLEDeviceScanResult>() {
            @Override
            public void onScanning(BLEDeviceScanResult result) {
                onStatusChange(STATUS_RECONNECTED);
                //打开设备
                device.open(result);
                //连接设备
                deviceConnect();
            }


            @Override
            public void onStop() {

            }

            @Override
            public void onError(DeviceCallbackException error) {
                //搜索设备失败
                onUpgradeError(error);
            }
        });
    }

    protected void deviceConnect() {
        device.connect(new DeviceConnectCallback() {
            @Override
            public void onSucceed(Device device, DeviceConnectResult result2) {
                //连接成功调用升级成功方法
                onUpgradeSuccess();
            }

            @Override
            public void onError(Device device, DeviceCallbackException error) {
                //设备连接失败，升级失败
                onUpgradeError(error);
            }
        });
    }


    public Device getDevice() {
        return device;
    }

    public String getAddress() {
        return address;
    }

    public File getFile() {
        return file;
    }

    public String getDeviceId() {
        return deviceId;
    }

    public Context getContext() {
        return context;
    }

    /**
     * 是否传输成功
     *
     * @return true 传输成功
     */
    public boolean isTransmitSuccess() {
        return isTransmitSuccess;
    }

    public boolean isTransmiting() {
        return isTransmiting;
    }
}
