package hk.advanpro.android.sdk.device.ota.advanpro;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.RandomAccessFile;

import hk.advanpro.android.sdk.commons.ALog;
import hk.advanpro.android.sdk.device.Device;
import hk.advanpro.android.sdk.device.ble.BLEBeltDevice;
import hk.advanpro.android.sdk.device.ble.BLEDevice;
import hk.advanpro.android.sdk.device.callback.DeviceCallbackException;
import hk.advanpro.android.sdk.device.callback.DeviceCommandCallback;
import hk.advanpro.android.sdk.device.callback.DeviceDisconnectCallback;
import hk.advanpro.android.sdk.device.enumeration.DeviceMode;
import hk.advanpro.android.sdk.device.ota.BLEFirmwareUpgradeBase;
import hk.advanpro.android.sdk.device.params.ble.belt.BLEBeltIntoBootloaderCommandParams;
import hk.advanpro.android.sdk.device.result.ble.belt.BLEBeltIntoBootloaderCommandResult;
import hk.advanpro.android.sdk.utils.CommonUtils;

/**
 * 呼吸带 固件升级
 * Created by jalin on 2017/12/15.
 */

public class BLEAdvanproFirmwareUpgrade extends BLEFirmwareUpgradeBase {
    public static final String TAG = "BLEAdvanproFirmwareUpgrade";

    public static final int UUID_Battery_Service = 0x180F;
    public static final int UUID_BELT_SERVERS = 0x8001;
    public static final int SMART_BELT_INFPIPE_Characteristic = 0x808F;
    public static final int Battery_Level_Characteristic = 0x2A19;

    private BLEConnect connB;
    private RandomAccessFile fi;
    private int upgradeState = 0;
    private int sectionLen = 0;
    private long progressTotal = 0;
    private long progress = 0;
    private int delay = 0;  // 蓝牙发送延迟，减少手机丢包重传

    @Override
    public void prepareUpgrade() {
        //准备固件升级
        //发送状态改变事件，进入ota
        onStatusChange(STATUS_INTO_BOOTLOADER);
        //先进入oat模式
        getDevice().command(BLEBeltDevice.CMD_BELT_INTO_BOOTLOADER, new BLEBeltIntoBootloaderCommandParams(), new DeviceCommandCallback<BLEBeltIntoBootloaderCommandResult>() {
            @Override
            public void onSuccess(Device device, BLEBeltIntoBootloaderCommandResult result) {
                if (!device.isClose()) {
                    //没有断开连接，直接断开连接
                    disconnect();
                } else {
                    //断开连接后，开始升级
                    startUpgrade();
                }
            }

            @Override
            public void onError(Device device, DeviceCallbackException error) {
                onUpgradeError(error);
            }
        });
    }

    private void disconnect() {
        getDevice().disconnect(new DeviceDisconnectCallback() {
            @Override
            public void onSucceed(Device device) {
                //断开连接后，开始升级
                startUpgrade();
            }

            @Override
            public void onError(Device device, DeviceCallbackException error) {
                onUpgradeError(error);
            }
        });
    }

    private void startUpgrade() {
        this.connB = null;
        try {
            this.fi = new RandomAccessFile(getFile(), "r");
        } catch (FileNotFoundException e) {
            e.printStackTrace();
        }
        this.progressTotal = this.getFile().length();
        this.progress = 0;
        this.handlerUpgrade.open();
        this.handlerUpgrade.sendEmptyMessage(MSG_CHECK_CONNECT);
    }


    private void success() {
        connB.disconnect();
        closefi();
        handlerUpgrade.close();
        ALog.d(TAG, "upgrade success");
        onTransmitComplete();
    }

    private void error(int e) {
        // 升级失败
        closefi();
        handlerUpgrade.close();
        ALog.d(TAG, "upgrade failed：" + e);
        upgradeState = -1;
        onUpgradeError(new DeviceCallbackException(e));
    }

    private void closefi() {
        try {
            fi.close();
        } catch (Exception e) {
        }
    }

    private void writePackage(byte[] data, int len) {
        try {
            Thread.sleep(delay);  // 每次发送前延迟
        } catch (InterruptedException e) {
            ALog.e(TAG, e);
        }
        connB.writeCharacteristic(UUID_BELT_SERVERS, SMART_BELT_INFPIPE_Characteristic, data);
        progress += len;
        if (progress > this.progressTotal)
            progress = this.progressTotal;
        onTransmitProgress(progressTotal, progress);
    }

    private void processSend() {
        try {
            if (progress < progressTotal) {
                switch (upgradeState) {
                    case 1: {
                        // 发送段头，6字节
                        upgradeState = 0x81;
                        byte[] sectionHead = new byte[6];
                        if (fi.read(sectionHead) == 6) {
                            sectionLen = (sectionHead[4] & 0xff) + (sectionHead[5] & 0xff) * 256;
                            ALog.d(TAG, "section head send:len=" + String.valueOf(sectionLen) + ", " + CommonUtils.bytesToHexString(sectionHead));
                            if (sectionLen <= 0) {
                                error(Device.ERROR_UPGRADE_OTHER_ERROR);
                            } else {
                                writePackage(sectionHead, 6);
                                handlerUpgrade.sendEmptyMessageDelayed(MSG_HEAD_DATA_TIMEOUT, 3000);
                            }
                        }
                        break;
                    }
                    case 2: {
                        if (sectionLen > 0) {
                            // 发送段内容，每次20字节
                            int len = 0;
                            byte[] sectionData = new byte[20];
                            if (sectionLen > 20) {
                                len = fi.read(sectionData, 0, 20);
                                sectionLen -= 20;
                            } else {
                                len = fi.read(sectionData, 0, sectionLen);
                                sectionLen = 0;
                            }
                            ALog.d(TAG, "section data send:more=" + String.valueOf(sectionLen) + ", " + CommonUtils.bytesToHexString(sectionData));
                            writePackage(sectionData, len);
                        }
                        if (sectionLen <= 0) {
                            // 当段数据发送完后，等待硬件回复2,3，如果3秒后还未收到，查询升级状态并进行整段重传
                            handlerUpgrade.sendEmptyMessageDelayed(MSG_CHECK_RECEIVE_STATUS, 3000);
                            upgradeState = 0x82;
                        }
                        break;
                    }
                }
            }
        } catch (Exception e) {
            ALog.e(TAG, e);
            error(Device.ERROR_UPGRADE_OTHER_ERROR);
        }
    }

    private void processRecv(byte[] data) {
        switch (data[0] & 0xff) {
            case 1:
                // 对段头的回复
                handlerUpgrade.removeMessages(MSG_HEAD_DATA_TIMEOUT);
                ALog.d(TAG, "section head receive ok:" + CommonUtils.bytesToHexString(data));
                if (data[1] == 0)
                    upgradeState = 2;
                else
                    error(Device.ERROR_UPGRADE_OTHER_ERROR);
                break;
            case 2:
                // 表示设备已接收完段数据
                ALog.d(TAG, "section data receive ok:" + CommonUtils.bytesToHexString(data));
                break;
            case 3:
                // 表示设备已写入段数据
                handlerUpgrade.removeMessages(MSG_CHECK_RECEIVE_STATUS);
                ALog.d(TAG, "section data write ok:" + CommonUtils.bytesToHexString(data));
                if (progress >= progressTotal) {
                    //发送状态改变事件，退出ota
                    onStatusChange(STATUS_OUT_BOOTLOADER);
                    // 升级成功，再次重启，进入应用模式
                    connB.writeCharacteristic(UUID_BELT_SERVERS, SMART_BELT_INFPIPE_Characteristic, new byte[]{0x04});
                    handlerUpgrade.sendEmptyMessageDelayed(MSG_UPGRADE_SUCCESS, 3000);  // 等待3秒
                } else {
                    upgradeState = 1;  // 发送下一段
                }
                break;
            case 0:
                // 设备段数据状态返回
                ALog.d(TAG, "section data received:" + CommonUtils.bytesToHexString(data));
                if (data.length == 11) {
                    int len = (data[7] & 0xff) + (data[8] & 0xff) * 256;  // 段长
                    int rec = (data[9] & 0xff) + (data[10] & 0xff) * 256;  // 已接收长度
                    if (rec < len) {
                        // 段没接收完，先清理，再处理重发，重发先发送段头
                        sectionLen = len;
                        progress -= sectionLen;
                        progress -= 6;
                        try {
                            fi.seek(fi.getFilePointer() - sectionLen - 6);
                            handlerUpgrade.removeMessages(MSG_CHECK_RECEIVE_STATUS);
                            handlerUpgrade.sendEmptyMessage(MSG_CLEAR_DATA);
                        } catch (IOException e) {
                            ALog.e(TAG, e);
                        }
                    }
                }
                break;
            case 0xff:
                // 段数据清理的回复
                ALog.d(TAG, "section data clear ok:" + CommonUtils.bytesToHexString(data));
                upgradeState = 1;
                break;
        }
    }

    public static final int MSG_CHECK_CONNECT = 9999;//连接通道异常,连接设备
    public static final int MSG_UPGRADE_SUCCESS = 8888;//升级成功
    public static final int MSG_DIS_CONNECT = 7777;//断开连接
    public static final int MSG_CHECK_RECEIVE_STATUS = 1111;//查询设备接收段数据状态
    public static final int MSG_CLEAR_DATA = 2222;//清理设备段数据
    public static final int MSG_HEAD_DATA_TIMEOUT = 3333;//段头发送超时

    private EventHandler handlerUpgrade = new EventHandler() {
        @Override
        public void handleMessage(android.os.Message msg) {
            switch (msg.what) {
                case MSG_CHECK_CONNECT: {
                    // 在启动10秒内，连接设备
                    // 如果连接通道异常，重连
                    if (connB == null || !connB.isCharacteristicEnable(SMART_BELT_INFPIPE_Characteristic)) {
                        onStatusChange(STATUS_RECONNECTED);
                        connB = BluetoothBLE.getInstance().connect(getContext(), getAddress(), this);
                    }
                    this.sendEmptyMessageDelayed(MSG_CHECK_CONNECT, 3000);
                    break;
                }

                case MSG_UPGRADE_SUCCESS: {
                    success();  // 升级成功
                    break;
                }

                case MSG_DIS_CONNECT: {
                    connB.disconnect();  // 断开连接，稍后会自动连接重连
                    onStatusChange(STATUS_DISCONNECTED);
                    break;
                }

                case MSG_CHECK_RECEIVE_STATUS: {
                    // 当接收超时时，查询设备接收段数据状态
                    if (delay < 100) delay += 5;  // 增加蓝牙发送延迟
                    connB.writeCharacteristic(UUID_BELT_SERVERS, SMART_BELT_INFPIPE_Characteristic, new byte[]{0x00});
                    ALog.d(TAG, "section data query:" + "00");
                    this.sendEmptyMessageDelayed(MSG_CHECK_RECEIVE_STATUS, 5000);
                    break;
                }

                case MSG_CLEAR_DATA: {
                    // 清理设备段数据
                    connB.writeCharacteristic(UUID_BELT_SERVERS, SMART_BELT_INFPIPE_Characteristic, new byte[]{-1});
                    ALog.d(TAG, "section data clear:" + "ff");
                    break;
                }

                case MSG_HEAD_DATA_TIMEOUT: {
                    // 段头发送超时
                    try {
                        upgradeState = 1;
                        progress -= 6;
                        fi.seek(fi.getFilePointer() - 6);
                        connB.disconnect();
                        onStatusChange(STATUS_RECONNECTED);
                        connB = BluetoothBLE.getInstance().connect(getContext(), getAddress(), this);
                    } catch (IOException e) {
                        ALog.e(TAG, e);
                    }
                    break;
                }

                case BluetoothBLE.MESSAGE_GATT_CONNECTED: {
                    try {
                        if (connB.existCharacteristic(UUID_BELT_SERVERS, Battery_Level_Characteristic) ||
                                connB.existCharacteristic(UUID_Battery_Service, Battery_Level_Characteristic)) {
                            //发送状态改变事件，进入ota
                            onStatusChange(STATUS_INTO_BOOTLOADER);
                            // 还处在应用模式，发送重启命令，进入bootloader模式
                            connB.writeCharacteristic(UUID_BELT_SERVERS, SMART_BELT_INFPIPE_Characteristic, new byte[]{0x04});
                            this.sendEmptyMessageDelayed(MSG_DIS_CONNECT, 2000);  // 等待2秒后断开并重连
                        } else {
                            // bootloader模式，开启升级通道
                            upgradeState = 1;
                            connB.setCharacteristicNotification(UUID_BELT_SERVERS, SMART_BELT_INFPIPE_Characteristic, true);
                            onTransmitProgress(progressTotal, progress);
                        }
                    } catch (Exception e) {
                        ALog.e(TAG, e);
                    }
                    break;
                }

                case BluetoothBLE.MESSAGE_DATA_REQUEST: {
                    if (connB.isCharacteristicEnable(SMART_BELT_INFPIPE_Characteristic)) {
                        processSend();
                    }
                    break;
                }

                case BluetoothBLE.MESSAGE_DATA_AVAILABLE: {
                    BLEData data = (BLEData) msg.obj;
                    if (data.value.length > 0 && data.characteristic == SMART_BELT_INFPIPE_Characteristic) {
                        processRecv(data.value);
                    }
                    break;
                }

                case BluetoothBLE.MESSAGE_GATT_DISCONNECTED:
                    break;
            }
        }
    };

    @Override
    protected void deviceScan() {

        //固件断开后，蓝牙广播不能及时开启，直接跳过扫描步骤
        //直接改为app状态
        if (getDevice() instanceof BLEBeltDevice)
            ((BLEBeltDevice) getDevice()).setInfo(BLEDevice.BLE_DEVICE_INFO_MODE, String.valueOf(DeviceMode.Application));

        //直接跳到连接步骤
        deviceConnect();
    }
}
