package hk.advanpro.android.sdk.device.ota.advanpro;

import android.bluetooth.BluetoothDevice;
import android.bluetooth.BluetoothGatt;
import android.bluetooth.BluetoothGattCallback;
import android.bluetooth.BluetoothGattCharacteristic;
import android.bluetooth.BluetoothGattDescriptor;
import android.bluetooth.BluetoothGattService;
import android.bluetooth.BluetoothProfile;
import android.content.Context;
import android.os.Message;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.UUID;

import hk.advanpro.android.sdk.commons.ALog;
import hk.advanpro.android.sdk.utils.CommonUtils;

/**
 * 蓝牙连接实例
 */
public class BLEConnect {
    public static final String TAG = "BLEConnect";

    private static final UUID CLIENT_CHARACTERISTIC_CONFIG = UUID.fromString("00002902-0000-1000-8000-00805f9b34fb");
    private BluetoothDevice mBluetoothDevice;
    private BluetoothGatt mBluetoothGatt;
    private EventHandler handler = null;

    private BLEConnect() {
    }

    /**
     * 创建连接
     * @param context context
     * @param device device
     * @param handler handler
     * @return BLEConnect
     */
    public static BLEConnect createConnect(Context context, BluetoothDevice device, EventHandler handler) {
        BLEConnect conn = new BLEConnect();
        conn.handler = handler;
        conn.mBluetoothDevice = device;
        conn.mBluetoothGatt = conn.mBluetoothDevice.connectGatt(context, false, conn.siGattCallBack);
        if (conn.mBluetoothGatt == null) {
            ALog.e(TAG, "Trying to create a new connection failed.");
            return null;
        }
        return conn;
    }

    /**
     * 断开连接
     */
    public void disconnect() {
        try {
            handlerBluetooth.close();
            if (mBluetoothGatt != null) {
                mBluetoothGatt.disconnect();
                refreshGatt();
                mBluetoothGatt.close();
                notifys.clear();
                reqList.clear();
            }
        } catch (Exception e) {
            ALog.e(TAG, e);
        }
    }

    /**
     * 是否空闲
     * @return boolean
     */
    public synchronized boolean isIdle() {
        return reqList.isEmpty();
    }

    /**
     * 读取Rssi信号强度
     * @return boolean
     */
    public boolean readRemoteRssi() {
        runCmd(new RssiCmd());
        return true;
    }

    /**
     * 读特征值
     *
     * @param service        服务UUID
     * @param characteristic 特征字
     * @return boolean
     */
    public boolean readCharacteristic(int service, int characteristic) {
        BluetoothGattCharacteristic c = getCharacteristic(service, characteristic);
        if (c == null) {
            ALog.e(TAG, "BluetoothGatt Service is not include Characteristic" + Integer.toHexString(service));
            return false;
        }

        runCmd(new ReadCmd(c));
        return true;
    }

    /**
     * 写特征值
     * @param service service
     * @param characteristic characteristic
     * @param value value
     * @return boolean
     */
    public boolean writeCharacteristic(int service, int characteristic, byte[] value) {
        BluetoothGattCharacteristic c = getCharacteristic(service, characteristic);
        if (c == null) {
            ALog.e(TAG, "BluetoothGatt Service is not include Characteristic " + Integer.toHexString(characteristic));
            return false;
        }

        runCmd(new WriteCmd(c, value));
        return true;
    }

    /**
     * 设置notification
     * @param service service
     * @param characteristic characteristic
     * @param enabled enabled
     * @return boolean
     */
    public boolean setCharacteristicNotification(int service, int characteristic, boolean enabled) {
        BluetoothGattCharacteristic c = getCharacteristic(service, characteristic);
        if (c == null) {
            ALog.e(TAG, "BluetoothGatt Service is not include Characteristic " + Integer.toHexString(characteristic));
            return false;
        }

        runCmd(new SetNotifyCmd(c, enabled));
        return true;
    }

    /**
     * 设置indication
     * @param service service
     * @param characteristic characteristic
     * @param enabled enabled
     * @return boolean
     */
    public boolean setCharacteristicIndication(int service, int characteristic, boolean enabled) {
        BluetoothGattCharacteristic c = getCharacteristic(service, characteristic);
        if (c == null) {
            ALog.e(TAG, "BluetoothGatt Service is not include Characteristic " + Integer.toHexString(characteristic));
            return false;
        }

        runCmd(new SetIndicationCmd(c, enabled));
        return true;
    }

    /**
     * 查询某特征字是否已开启notification/indication
     * @param characteristic characteristic
     * @return boolean
     */
    public boolean isCharacteristicEnable(int characteristic) {
        return (notifys.contains(characteristic));
    }

    /**
     * 是否存在服务是否存在
     * @param service service
     * @return boolean
     */
    public boolean existService(int service) {
        return getService(service) != null;
    }

    /**
     * 查询特征字是否存在
     * @param service service
     * @param characteristic characteristic
     * @return boolean
     */
    public boolean existCharacteristic(int service, int characteristic) {
        return getCharacteristic(service, characteristic) != null;
    }

    private int getShortUUID(UUID uuid) {
        String str = uuid.toString().substring(4, 8);
        return Integer.parseInt(str, 16);
    }

    private BluetoothGattService getService(int service) {
        List<BluetoothGattService> list = mBluetoothGatt.getServices();
        if (list != null) {
            for (BluetoothGattService s : list) {
                int n2 = getShortUUID(s.getUuid());
                if (service == 0 || service == n2) {
                    return s;
                }
            }
        }
        return null;
    }

    private BluetoothGattCharacteristic getCharacteristic(BluetoothGattService service, int characteristic) {
        List<BluetoothGattCharacteristic> list = service.getCharacteristics();
        if (list != null) {
            for (BluetoothGattCharacteristic c : list) {
                int n2 = getShortUUID(c.getUuid());
                if (characteristic == 0 || characteristic == n2) {
                    return c;
                }
            }
        }
        return null;
    }

    private void refreshGatt() {
        try {
            Method localMethod = mBluetoothGatt.getClass().getMethod("refresh");
            if (localMethod != null) {
                localMethod.invoke(mBluetoothGatt);
            }
        } catch (Exception localException) {
            ALog.d(TAG, "An exception occurred while refreshing device");
        }
    }

    private BluetoothGattCharacteristic getCharacteristic(int service, int characteristic) {
        if (mBluetoothGatt == null) {
            ALog.e(TAG, "BluetoothAdapter  not initialized or BluetoothGatt is null");
            return null;
        }

        BluetoothGattService Gattservice = getService(service);
        if (Gattservice == null) {
            ALog.e(TAG, "BluetoothGatt Service is not include SERVICE " + Integer.toHexString(service));
            return null;
        }

        return getCharacteristic(Gattservice, characteristic);
    }

    private final ArrayList<Integer> notifys = new ArrayList<>();
    private final BluetoothGattCallback siGattCallBack = new BluetoothGattCallback() {
        @Override
        public void onConnectionStateChange(BluetoothGatt gatt, int status, int newState) {
            if (newState == BluetoothProfile.STATE_CONNECTED) {
                ALog.d(TAG, "Connected to GATT server.");
                notifys.clear();
                boolean bDisSerSuceed = mBluetoothGatt.discoverServices();  // 搜索支持的服务
                ALog.d(TAG, "Attempting to start service discovery:" + bDisSerSuceed);
            } else if (newState == BluetoothProfile.STATE_DISCONNECTED) {
                ALog.d(TAG, "Disconnected from GATT server.");
                mBluetoothGatt.close();
                notifys.clear();
                refreshGatt();
                handler.sendEmptyMessage(BluetoothBLE.MESSAGE_GATT_DISCONNECTED);
                handlerBluetooth.close();
            }
        }

        @Override
        public void onServicesDiscovered(BluetoothGatt gatt, int status) {
            ALog.d(TAG, "onServicesDiscovered");
            handler.sendEmptyMessage(BluetoothBLE.MESSAGE_GATT_CONNECTED);
            handlerBluetooth.open();
        }

        @Override
        public void onReadRemoteRssi(BluetoothGatt gatt, int rssi, int status) {
            // 获取到RSSI，RSSI 正常情况下是一个负值，如-33；这个值的绝对值越小，代表设备离手机越近
            // 通过mBluetoothGatt.readRemoteRssi();来获取
            Message msg = new Message();
            msg.what = BluetoothBLE.MESSAGE_RSSI;
            msg.obj = rssi;
            handler.sendMessage(msg);
            runNextCmd();
        }

        @Override
        public void onCharacteristicRead(BluetoothGatt gatt, final BluetoothGattCharacteristic characteristic, int status) {
            // 读取到值
            if (status == BluetoothGatt.GATT_SUCCESS) {
                Message msg = new Message();
                msg.what = BluetoothBLE.MESSAGE_DATA_AVAILABLE;
                msg.obj = new BLEData(
                        getShortUUID(characteristic.getService().getUuid()),
                        getShortUUID(characteristic.getUuid()),
                        characteristic.getValue());
                handler.sendMessage(msg);
            }
            runNextCmd();
        }

        @Override
        public void onCharacteristicWrite(BluetoothGatt gatt, BluetoothGattCharacteristic characteristic, int status) {
            if (status == BluetoothGatt.GATT_SUCCESS) {
                // write成功
            } else {
                // write失败，重传
                ALog.d(TAG, "onCharacteristicWrite error");
            }
            runNextCmd();
        }

        @Override
        public void onDescriptorWrite(BluetoothGatt gatt, BluetoothGattDescriptor descriptor, int status) {
            int characteristic = getShortUUID(descriptor.getCharacteristic().getUuid());
            if (!notifys.contains(characteristic)) {
                notifys.add(characteristic);
            }
            runNextCmd();
        }

        @Override
        public void onCharacteristicChanged(BluetoothGatt gatt, BluetoothGattCharacteristic characteristic) {
            // 收到设备notify值 （设备上报值）
            Message msg = new Message();
            msg.what = BluetoothBLE.MESSAGE_DATA_AVAILABLE;
            msg.obj = new BLEData(
                    getShortUUID(characteristic.getService().getUuid()),
                    getShortUUID(characteristic.getUuid()),
                    characteristic.getValue());
            handler.sendMessage(msg);
        }
    };

    private final LinkedList<Runnable> reqList = new LinkedList<>();

    private void runCmd(Runnable cmd) {
        boolean b = false;
        synchronized (reqList) {
            b = reqList.isEmpty();
            reqList.add(cmd);
        }
        if (b) cmd.run();
    }

    private void runNextCmd() {
        Runnable runnable = null;
        synchronized (reqList) {
            if (!reqList.isEmpty()) {
                reqList.pop();
            }
            if (reqNextCmd() > 0) {
                runnable = reqList.getFirst();
            }
        }
        if (runnable != null) {
            runnable.run();
        }
    }

    private int reqNextCmd() {
        synchronized (reqList) {
            if (reqList.isEmpty()) {
                handler.sendEmptyMessage(BluetoothBLE.MESSAGE_DATA_REQUEST);
            }
            return reqList.size();
        }
    }

    private final EventHandler handlerBluetooth = new EventHandler() {
        @Override
        public void open() {
            super.open();
            this.sendEmptyMessage(1);
        }

        @Override
        public void close() {
            super.close();
            this.removeMessages(1);
        }

        @Override
        public void handleMessage(Message msg) {
            reqNextCmd();
            sendEmptyMessageDelayed(1, 100);
        }
    };

    //-----------------------------------------------------------------
    private class RssiCmd implements Runnable {
        @Override
        public void run() {
            mBluetoothGatt.readRemoteRssi();
        }
    }

    //-----------------------------------------------------------------
    private class ReadCmd implements Runnable {
        BluetoothGattCharacteristic c;

        public ReadCmd(BluetoothGattCharacteristic c) {
            this.c = c;
        }

        @Override
        public void run() {
            mBluetoothGatt.readCharacteristic(c);
        }
    }

    //-----------------------------------------------------------------
    private class WriteCmd implements Runnable {
        BluetoothGattCharacteristic c;
        byte[] value;

        public WriteCmd(BluetoothGattCharacteristic c, byte[] value) {
            this.c = c;
            this.value = value;
        }

        @Override
        public void run() {
            c.setValue(value);
            mBluetoothGatt.writeCharacteristic(c);
            ALog.d(TAG, "write:" + CommonUtils.bytesToHexString(value));
        }
    }

    //-----------------------------------------------------------------
    private class SetNotifyCmd implements Runnable {
        BluetoothGattCharacteristic c;
        boolean enable;

        public SetNotifyCmd(BluetoothGattCharacteristic c, boolean enable) {
            this.c = c;
            this.enable = enable;
        }

        @Override
        public void run() {
            boolean succeed = mBluetoothGatt.setCharacteristicNotification(c, enable);
            if (!succeed) {
                ALog.e(TAG, "failed to notify the Characteris " + c.getUuid().toString());
                return;
            }

            BluetoothGattDescriptor descriptor = c.getDescriptor(CLIENT_CHARACTERISTIC_CONFIG);
            if (descriptor == null) {
                return;
            }

            if (enable)
                descriptor.setValue(BluetoothGattDescriptor.ENABLE_NOTIFICATION_VALUE);
            else
                descriptor.setValue(BluetoothGattDescriptor.DISABLE_NOTIFICATION_VALUE);

            mBluetoothGatt.writeDescriptor(descriptor);
        }
    }

    //-----------------------------------------------------------------
    private class SetIndicationCmd implements Runnable {
        BluetoothGattCharacteristic c;
        boolean enable;

        public SetIndicationCmd(BluetoothGattCharacteristic c, boolean enable) {
            this.c = c;
            this.enable = enable;
        }

        @Override
        public void run() {
            boolean succeed = mBluetoothGatt.setCharacteristicNotification(c, enable);
            if (!succeed) {
                ALog.e(TAG, "failed to notify the Characteris " + c.getUuid().toString());
                return;
            }

            BluetoothGattDescriptor descriptor = c.getDescriptor(CLIENT_CHARACTERISTIC_CONFIG);
            if (descriptor == null) {
                return;
            }

            if (enable)
                descriptor.setValue(BluetoothGattDescriptor.ENABLE_INDICATION_VALUE);
            else
                descriptor.setValue(BluetoothGattDescriptor.DISABLE_NOTIFICATION_VALUE);

            mBluetoothGatt.writeDescriptor(descriptor);
        }
    }
}