/******************************************************************************
 *  Copyright (C) Cambridge Silicon Radio Limited 2014
 *
 *  This software is provided to the customer for evaluation
 *  purposes only and, as such early feedback on performance and operation
 *  is anticipated. The software source code is subject to change and
 *  not intended for production. Use of developmental release software is
 *  at the user's own risk. This software is provided "as is," and CSR
 *  cautions users to determine for themselves the suitability of using the
 *  beta release version of this software. CSR makes no warranty or
 *  representation whatsoever of merchantability or fitness of the product
 *  for any particular purpose or use. In no event shall CSR be liable for
 *  any consequential, incidental or special damages whatsoever arising out
 *  of the use of or inability to use this software, even if the user has
 *  advised CSR of the possibility of such damages.
 *
 ******************************************************************************/

package hk.advanpro.android.sdk.device.ota.csr;

import android.os.Bundle;
import android.os.Handler;
import android.os.Looper;
import android.os.Message;
import android.os.ParcelUuid;
import android.util.Log;

import java.lang.ref.WeakReference;

class ConnectionHandler extends Handler {
    private String mac;
    private final static byte UNKNOWN_SERVICE_MODE = 0x0;
    private final static byte APPLICATION_SERVICE_MODE = 0x01;
    private final static byte BOOTLOADER_SERVICE_MODE = 0x02;
    private final static byte REFRESH_SERVICE_MODE = 0x03;

    /*
     * Constructor creates a StateMachine in main thread.
     */
    protected ConnectionHandler(String mac) {
        super(Looper.getMainLooper());
        this.mac = mac;
    }

    @Override
    public void handleMessage(Message msg) {
        IOtauMessageListener listener = null;
        WeakReference<OtauManager> weakRef = new WeakReference<>(OtauManager.getInstance(mac));
        if (weakRef.get() != null)
            listener = weakRef.get().getListener();

        Log.d("ConnectionHandler", "State.getOtaState() = " + State.getOtaState(mac) + ", State.getState() = " + State.getState(mac));
        switch (msg.what) {
            case OtauClient.MESSAGE_CONNECTED: {
                // csr added start
                // prevent from to get discover completed event twice very quickly
                if (weakRef.get() != null)
                    weakRef.get().clearClientRequest();
                else
                    return;
                try {
                    Thread.sleep(1000);
                } catch (InterruptedException e) {
                    Log.e("ConnectionHandler", "InterruptedException!!!");
                }
                State.setState(mac, State.ConnectionState.STATE_CONNECTED);
                State.setOtaState(mac, State.OtaState.STATE_OTA_IDLE);
                // csr added end

                byte mode = checkServices(msg);
                if (mode == APPLICATION_SERVICE_MODE) {
                    Log.d("ConnectionHandler", "in application mode");
                    handleApplicationService();
                } else if (mode == BOOTLOADER_SERVICE_MODE) {
                    Log.d("ConnectionHandler", "in bootloader mode");
                    handleBootloaderService();
                } else if (mode == REFRESH_SERVICE_MODE) {
                    Log.d("ConnectionHandler", "Refresh Service mode");
                    break;
                } else {
                    State.setState(mac, State.ConnectionState.STATE_INVALID);
                    Log.d("ConnectionHandler", "in unknown mode");
                    if (listener != null)
                        listener.onConnectionStateUpdate(-1);
                }
            }
            break;
            case OtauClient.MESSAGE_DISCONNECTED: {
                Log.d("ConnectionHandler", "MESSAGE_DISCONNECTED");
                if (weakRef.get() != null) {
                    weakRef.get().disconnect();
                    if (State.getOtaState(mac) == State.OtaState.STATE_OTA_SET_MODE) {
                        Log.d("ConnectionHandler", "reconnect for boot mode");
                        weakRef.get().connect();
                    } else if (State.getOtaState(mac) == State.OtaState.STATE_OTA_SET_TRANSFER_COMPLETE) {
                        if (weakRef.get().getIsMeshDevice()) {
                            Log.d("ConnectionHandler", "reconnect for refreshing");
                            weakRef.get().connect(true);
                        }
                    }
                }

                if (listener != null)
                    listener.onConnectionStateUpdate(OtauClient.MESSAGE_DISCONNECTED);
            }
            break;
        }

    }

    /*
     * Handle Application service attributes
     *
     */
    private synchronized void handleApplicationService() {
        IOtauMessageListener listener = null;
        WeakReference<OtauManager> weakRef = new WeakReference<>(OtauManager.getInstance(mac));
        if (weakRef.get() != null)
            listener = weakRef.get().getListener();

        State.setState(mac, State.ConnectionState.STATE_CONNECTED);
        // csr remove start
        //State.setOtaState(State.OtaState.STATE_OTA_IDLE);
        // csr remove end

        if (weakRef.get() != null) {

            /* After refreshing service attributes.
             * It read new software version and trigger pairing and disconnected.
             */
            if (weakRef.get().getRefreshAttributesDone()) {
                weakRef.get().setRefreshAttributesDone(false);
                weakRef.get().getSoftwareVersion();
                weakRef.get().readAuthenticatedCharacteristicInApp();
                State.setOtaState(mac, State.OtaState.STATE_OTA_REFRESH_ATTRIBUTES);
                return;
            }

            if (listener != null)
                listener.onConnectionStateUpdate(OtauClient.MESSAGE_CONNECTED);

            Log.d("ConnectionHandler", "getSoftwareVersion");
            weakRef.get().getSoftwareVersion();
            Log.d("ConnectionHandler", "readAuthenticatedCharacteristicInApp");
            weakRef.get().readAuthenticatedCharacteristicInApp();
            Log.d("ConnectionHandler", "registerServiceChangedIndication");
            weakRef.get().registerServiceChangedIndication();
            weakRef.get().registerDataTransferNotification(OtauClient.BtSmartUuid.OTAU_APPLICATION_SERVICE.getUuid());

            if (!weakRef.get().getCsBlockSupported()) {   // Bootloader version 4
                Log.d("ConnectionHandler", "readCSKeys for Bootloader version 4");
                weakRef.get().readCSKeys(OtauManager.REQUEST_READ_CSKEYS_BT, OtauManager.USER_KEY_BT_ADDRESS);
                weakRef.get().readCSKeys(OtauManager.REQUEST_READ_CSKEYS_XLST, OtauManager.USER_KEY_XTAL_TRIM);
                weakRef.get().readCSKeys(OtauManager.REQUEST_READ_CSKEYS_IDROOT, OtauManager.USER_KEY_IDENTITY_ROOT);
                weakRef.get().readCSKeys(OtauManager.REQUEST_READ_CSKEYS_ENCROOT, OtauManager.USER_KEY_ENCRYPTION_ROOT);
                weakRef.get().enableOTAMode();
            } else {
                Log.d("ConnectionHandler", "readNextCsBlock");
                weakRef.get().readNextCsBlock();
            }
        }
    }

    /*
     * Handle Bootloader service attributes
     *
     */
    private synchronized void handleBootloaderService() {
        IOtauMessageListener listener = null;
        WeakReference<OtauManager> weakRef = new WeakReference<>(OtauManager.getInstance(mac));
        if (weakRef.get() != null)
            listener = weakRef.get().getListener();

        if (listener != null)
            listener.onConnectionStateUpdate(OtauClient.MESSAGE_CONNECTED);

        if (weakRef.get() != null) {

            // Handle case for V4/V5
            if (!weakRef.get().getBootloaderSoftwareVersionSupported()) {
                Log.d("ConnectionHandler", "Handle case for V4/V5");

                if (weakRef.get().getHostValidation()) {
                    weakRef.get().readDataTransferCharacteristic();
                } else {
                    weakRef.get().registerBootloaderTransferControlNotifiocation();
                    weakRef.get().readBootloadVersion();

                    // This process is for V5 onwards
                    if (weakRef.get().getReadCskeyFromBootloader()) {
                        weakRef.get().readCSKeysFromBootloaderService(OtauManager.REQUEST_READ_CSKEYS_BT_FROM_BL,
                                OtauManager.USER_KEY_BT_ADDRESS);
                        weakRef.get().readDataTransferCharacteristic();
                        weakRef.get().readCSKeysFromBootloaderService(OtauManager.REQUEST_READ_CSKEYS_XLST_FROM_BL,
                                OtauManager.USER_KEY_XTAL_TRIM);
                        weakRef.get().readDataTransferCharacteristic();
                        weakRef.get().readCSKeysFromBootloaderService(OtauManager.REQUEST_READ_CSKEYS_IDROOT_FROM_BL,
                                OtauManager.USER_KEY_IDENTITY_ROOT);
                        weakRef.get().readDataTransferCharacteristic();
                        weakRef.get().readCSKeysFromBootloaderService(OtauManager.REQUEST_READ_CSKEYS_ENCROOT_FROM_BL,
                                OtauManager.USER_KEY_ENCRYPTION_ROOT);
                        weakRef.get().readDataTransferCharacteristic();
                    }

                    if (listener != null)
                        listener.onOtauStarted();
                }
            } else {
                /*This is for non-bonding on-chip applications to trigger temporary pairing*/
                if (!weakRef.get().isBonded()) {
                    weakRef.get().readAuthenticatedCharacteristic();
                    //weakRef.get().readTransferControlStatus();
                }
                // Auto detect challenge response setting
                Log.d("ConnectionHandler", "readTransferControlStatus");
                weakRef.get().readTransferControlStatus();
            }
        }
    }

    /*
     * Check services from
     *
     * @param msg
     *           The Android Message object which will be sent out
     */
    private synchronized byte checkServices(Message msg) {
        WeakReference<OtauManager> weakRef = new WeakReference<>(OtauManager.getInstance(mac));
        Bundle bundle = msg.getData();
        ParcelUuid[] ServiceUuidArray = (ParcelUuid[]) bundle.getParcelableArray(OtauClient.EXTRA_SERVICE_UUID_ARRAY);
        boolean isRefresh = bundle.getBoolean(OtauClient.EXTRA_REFRESH_ATTRIBUTES);
        boolean isRefreshDone = bundle.getBoolean(OtauClient.EXTRA_REFRESH_ATTRIBUTES_DONE);
        weakRef.get().setCsBlockSupported(false);
        weakRef.get().setReadCskeyFromBootloader(false);
        weakRef.get().setReadBootloaderSoftwareVersion(false);
        weakRef.get().setRefreshAttributesDone(isRefreshDone);
        Log.d("ConnectionHandler", "checkServices: isRefresh = " + isRefresh);
        if (isRefresh) {
            weakRef.get().refreshAttribute();
            return REFRESH_SERVICE_MODE;
        }

        for (ParcelUuid uuid : ServiceUuidArray) {
            if (uuid.getUuid().compareTo(OtauClient.BtSmartUuid.OTAU_APPLICATION_SERVICE.getUuid()) == 0) {
                Log.d("ConnectionHandler", "found application UUID");
                ParcelUuid[] CharaUuidArray = (ParcelUuid[]) bundle.getParcelableArray(OtauClient.EXTRA_CHARACTERISTIC_UUID_ARRAY);
                if (CharaUuidArray != null) {
                    for (ParcelUuid charaUuid : CharaUuidArray) {
                        Log.d("ConnectionHandler", "charaUuid = " + charaUuid.getUuid());
                        if (charaUuid.getUuid().compareTo(OtauClient.BtSmartUuid.OTAU_APPLICATION_READCSBLOCK.getUuid()) == 0) {
                            weakRef.get().setCsBlockSupported(true);
                        }
                        if (charaUuid.getUuid().compareTo(OtauClient.BtSmartUuid.OTAU_CSR_OTA_VERSION.getUuid()) == 0) {
                            weakRef.get().setCsrOtaVersion(true);
                        }
                    }
                }
                return APPLICATION_SERVICE_MODE;
            } else if (uuid.getUuid().compareTo(OtauClient.BtSmartUuid.OTAU_BOOTLOADER_SERVICE.getUuid()) == 0) {
                Log.d("ConnectionHandler", "found bootloader UUID");
                ParcelUuid[] CharaUuidArray = (ParcelUuid[]) bundle.getParcelableArray(OtauClient.EXTRA_CHARACTERISTIC_UUID_ARRAY);
                if (CharaUuidArray != null) {
                    for (ParcelUuid charaUuid : CharaUuidArray) {
                        Log.d("ConnectionHandler", "charaUuid = " + charaUuid.getUuid());
                        if (charaUuid.getUuid().compareTo(OtauClient.BtSmartUuid.OTAU_APPLICATION_READCSKEY.getUuid()) == 0) {
                            weakRef.get().setReadCskeyFromBootloader(true);
                        }
                        if (charaUuid.getUuid().compareTo(OtauClient.BtSmartUuid.OTAU_BOOTLOADER_SOFTWARE_VERSION.getUuid()) == 0) {
                            weakRef.get().setReadBootloaderSoftwareVersion(true);
                        }
                    }
                }
                return BOOTLOADER_SERVICE_MODE;
            }
        }

        return UNKNOWN_SERVICE_MODE;
    }
}
