/******************************************************************************
 *  Copyright (C) Cambridge Silicon Radio Limited 2014
 *
 *  This software is provided to the customer for evaluation
 *  purposes only and, as such early feedback on performance and operation
 *  is anticipated. The software source code is subject to change and
 *  not intended for production. Use of developmental release software is
 *  at the user's own risk. This software is provided "as is," and CSR
 *  cautions users to determine for themselves the suitability of using the
 *  beta release version of this software. CSR makes no warranty or
 *  representation whatsoever of merchantability or fitness of the product
 *  for any particular purpose or use. In no event shall CSR be liable for
 *  any consequential, incidental or special damages whatsoever arising out
 *  of the use of or inability to use this software, even if the user has
 *  advised CSR of the possibility of such damages.
 *
 ******************************************************************************/

package hk.advanpro.android.sdk.device.ota.csr;

import android.bluetooth.BluetoothAdapter;
import android.bluetooth.BluetoothDevice;
import android.bluetooth.BluetoothManager;
import android.content.Context;
import android.os.Handler;
import android.os.Looper;
import android.util.Log;

import java.io.UnsupportedEncodingException;
import java.security.GeneralSecurityException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.UUID;

import javax.crypto.Cipher;
import javax.crypto.SecretKey;
import javax.crypto.spec.SecretKeySpec;

public class OtauManager {
    // OTA request message id
    public static final int REQUEST_SERVICE_CHANGED = 1;
    public static final int REQUEST_DATATRANSFER_NOTIFICATION = 2;
    public static final int REQUEST_READ_CSKEYS_BT = 3;
    public static final int REQUEST_READ_CSKEYS_XLST = 4;
    public static final int REQUEST_READ_CSKEYS_IDROOT = 5;
    public static final int REQUEST_READ_CSKEYS_ENCROOT = 6;
    public static final int REQUEST_ENABLE_OTA_MODE = 7;
    public static final int REQUEST_BL_TRANSFERCONTROL_NOTIFICATION = 8;
    public static final int REQUEST_READ_DATA_TRANSFER = 9;
    public static final int REQUEST_WRITE_DATA_TRANSFER = 10;
    public static final int REQUEST_READ_BOOTLOADER_VERSION = 11;
    public static final int REQUEST_SET_APPLCATION_ONCHIP = 12;
    public static final int REQUEST_GET_APPLCATION_ONCHIP = 13;
    public static final int REQUEST_SET_TRANSFER_CONTROL_READY = 14;
    public static final int REQUEST_SET_TRANSFER_CONTROL_INPROGRESS = 15;
    public static final int REQUEST_WRITE_BYTES_TO_TARGET = 16;
    public static final int REQUEST_SET_TRANSFER_CONTROL_COMPLETE = 17;
    public static final int REQUEST_PAUSE_DOWNLOAD = 18;
    public static final int REQUEST_CANCEL_DOWNLOAD = 19;
    public static final int REQUEST_READ_CSBLOCK = 20;
    public static final int REQUEST_READ_CSKEYS_BT_FROM_BL = 21;
    public static final int REQUEST_READ_CSKEYS_XLST_FROM_BL = 22;
    public static final int REQUEST_READ_CSKEYS_IDROOT_FROM_BL = 23;
    public static final int REQUEST_READ_CSKEYS_ENCROOT_FROM_BL = 24;
    public static final int REQUEST_READ_TRANSFER_CONTROL_STATUS = 25;
    public static final int REQUEST_SET_TRANSFER_CONTROL_FAILED = 26;
    public static final int REQUEST_READ_BL_SOFTWARE_VERSION    = 27;
    public static final int REQUEST_READ_CSR_OTA_VERSION    = 28;
    public static final int REQUEST_READ_DEVICE_SOFTWARE_VERSION    = 29;
    public static final int REQUEST_READ_AUTHENTICATED_CHARACTERISTIC = 30;

    // These are boot loader transfer control value
    public static final short SET_TRANSFER_CONTROL_INIT = 0;
    public static final short SET_TRANSFER_CONTROL_READY = 1;
    public static final short SET_TRANSFER_CONTROL_INPROGRESS = 2;
    public static final short SET_TRANSFER_PAUSED = 3;
    public static final short SET_TRANSFER_COMPLETED = 4;
    public static final short SET_TRANSFER_FAILED = 5;
    public static final short SET_TRANSFER_ABORT = 6;

    // These are read CS keys value
    public final static int USER_KEY_BT_ADDRESS = 0x01;
    public final static int USER_KEY_XTAL_TRIM = 0x02;
    public final static int USER_KEY_IDENTITY_ROOT = 0x11;
    public final static int USER_KEY_ENCRYPTION_ROOT = 0x12;

    // These are error message ids
    public static final int ERROR_MESSAGE_ID_READCSKEY_BTADDRESS = 1;
    public static final int ERROR_MESSAGE_ID_READCSKEY_XLST = 2;
    public static final int ERROR_MESSAGE_ID_READCSKEY_IDROOT = 3;
    public static final int ERROR_MESSAGE_ID_READCSKEY_ENCROOT = 4;
    public static final int ERROR_MESSAGE_ID_CHALLENGE_RESPONSE = 5;
    public static final int ERROR_MESSAGE_ID_ENABLE_OTA = 6;
    public static final int ERROR_MESSAGE_ID_WRITE_TO_TARGET = 7;
    public static final int ERROR_MESSAGE_ID_BL_TRANSFERCONTROL_NOTIFICATION = 8;
    public static final int ERROR_MESSAGE_ID_NULL_CHARACTERISTIC = 9;
    public static final int ERROR_MESSAGE_ID_NULL_SERVICE = 10;
    public static final int ERROR_MESSAGE_ID_SET_TRANSFER_CONTROL_INPROGRESS = 11;
    public static final int ERROR_MESSAGE_ID_SET_APPLICATION_ID = 12;
    public static final int ERROR_MESSAGE_ID_OTHERS = 13;
    public static final int ERROR_MESSAGE_GATT_CLIENT_DESTROYED = 14;
    public static final int ERROR_MESSAGE_INIT_INTERRUPTED = 15;

    // These are CS Keys offset and size definition
    public static final byte READ_CSKEY_BUILDID_OFFSET = 0;
    public static final byte READ_CSKEY_BUILDID_LENGTH = 2;

    private final static int DATA_LENGTH = 20;
    private final static int CONNECTION_DELAY_TIME_MS      = 200;

    private static Map<String, OtauManager> managerMap = new HashMap<>();
    public OtauClient mGattClient = null;
    public BluetoothDevice mBLEDevice = null;
    private OtauMessageHandler mStatemachine;
    private ArrayList<CskeyItem> mCskeyList;
    private boolean mCsBlockSupported = false;
    private boolean mReadCskeyFromBootloader = false;
    private boolean mSoftwareVersionSupported = false;
    private boolean mCsrOtaVersionSupported = false;

    private int mProgressing = 0;
    private byte[] mImageData;
    private IOtauMessageListener mListener;
    private boolean isDownloadpaused = false;
    private int mSentNum = 0;
    private boolean isAesEncryptionEnabled = false;
    private ArrayList<CskeyItem> mMergedCsKeyList;
    private boolean mIsCancelDownload = false;
    private boolean mRefreshAttributesDone = false;
    private boolean mIsMeshdevice = false;

    static byte[] mSharedSecretKeySdk = new byte[] { (byte) 0x00, (byte) 0x11, (byte) 0x22, (byte) 0x33, (byte) 0x44,
            (byte) 0x55, (byte) 0x66, (byte) 0x77, (byte) 0x88, (byte) 0x99, (byte) 0xaa, (byte) 0xbb, (byte) 0xcc,
            (byte) 0xdd, (byte) 0xee, (byte) 0xff };

    /**
     * This is the handler for general connection messages.
     */
    private ConnectionHandler mDeviceHandler;

    /*
     * make default constructor invisible
     */
    private OtauManager() {
    }

    /*
     * Initialising OTA update manager
     * 
     * @param context Activity context
     */
    public static boolean initialize(Context context, BluetoothDevice device) {
        if (device == null) {
            return false;
        }
        BluetoothManager bluetoothManager = (BluetoothManager) context.getSystemService(Context.BLUETOOTH_SERVICE);
        if (bluetoothManager == null) {
            return false;
        }

        BluetoothAdapter bluetoothAdapter = bluetoothManager.getAdapter();
        if (bluetoothAdapter == null) {
            return false;
        }

        OtauManager manager = new OtauManager();
        manager.mStatemachine = new OtauMessageHandler(device.getAddress());
        manager.mDeviceHandler = new ConnectionHandler(device.getAddress());
        manager.mGattClient = new OtauClient(context, CONNECTION_DELAY_TIME_MS);
        manager.mBLEDevice = device;

        State.setState(device.getAddress(), State.ConnectionState.STATE_IDLE);
        State.setScanState(device.getAddress(), State.ScanState.STATE_SCAN_IDLE);
        State.setOtaState(device.getAddress(), State.OtaState.STATE_OTA_IDLE);
        State.setReadCsBlockState(device.getAddress(), State.ReadCsBlockState.READ_CS_BLOCK_BUILD_ID);
        manager.mMergedCsKeyList = new ArrayList<>();
        manager.mIsCancelDownload = false;
		managerMap.put(device.getAddress(), manager);
        return true;
    }

    public static void destroyAllManager() {
        for (OtauManager manager : managerMap.values()) {
            manager.mGattClient.destroy();
        }
        managerMap.clear();
    }
    
    /*
     * Get OTA update manage singleton instance
     */
    public static OtauManager getInstance(String mac) {
        return managerMap.get(mac);
    }

    /*
     * Destroy this instance
     * 
     * @param context Activity context
     */
    public void destroy() {
		mGattClient.destroy();
        managerMap.remove(mBLEDevice.getAddress());
    }
    
    /*
     * Host application acts as Gatt client to connect BLE device
     * 
     * @param device BluetoothDevice The BLE device instance
     * 
     * @param listener IOtauMessageListener The listener who are interested in OTA message
     */
    public void connect(IOtauMessageListener listener) {
        if (mBLEDevice != null && mDeviceHandler != null && mGattClient != null) {
        	// csr added start
        	State.setOtaState(mBLEDevice.getAddress(), State.OtaState.STATE_OTA_IDLE);
        	// csr added end
            mListener = listener;
            mGattClient.connectAsClient(mBLEDevice, mDeviceHandler);
        }
    }

    /*
     * Provide interface for reconnect to BLE devices after switching to bootloader mode
     */
    public void connect() {
        if (mBLEDevice != null && mDeviceHandler != null && mGattClient != null) {
            mGattClient.connectAsClient(mBLEDevice, mDeviceHandler);
        }
    }

    public void connect(final boolean refresh) {
        if (mBLEDevice != null && mDeviceHandler != null && mGattClient != null) {
            if(refresh){
                if(mListener != null){
                    mListener.onOtaStateUpdate(State.OtaState.STATE_OTA_RECONNECT_GATT);
                }
                new Handler(Looper.getMainLooper()).postDelayed(new Runnable() {
                    @Override
                    public void run() {
                        mGattClient.connectAsClient(mBLEDevice, mDeviceHandler, true);
                    }
                }, 1500);
            }
        }
    }

	/**
	 * 刷新服务
	 */
	public void refresh() {
		if (mGattClient != null) {
			mGattClient.refresh();
		}
	}
	
    /*
     * Disconnect BLE devices
     */
    public void disconnect() {
        if (mGattClient != null) {
            mGattClient.disconnect();
        }
    }

    /*
     * Disconnect BLE devices
     */
    public void disconnectAndClose() {
        if (mGattClient != null) {
            mGattClient.disconnectAndClose();
        }
    }
	
    /*
     * Tell host application if remote application support challenge response
     */
    public void setHostValidation(boolean enable) {
        isAesEncryptionEnabled = enable;
    }

    /*
     * Check if remote device supports challenge response
     */
    public boolean getHostValidation() {
        return isAesEncryptionEnabled;
    }

    IOtauMessageListener getListener() {
        return mListener;
    }

    /**
     * read characteristic to get Boot loader version
     */
    public void readBootloadVersion() {
        if (mGattClient != null) {
            if (mListener != null) {
                mListener.onOtherLog("readBootloadVersion...");
            }
            mGattClient.requestCharacteristicValue(REQUEST_READ_BOOTLOADER_VERSION,
                    OtauClient.BtSmartUuid.OTAU_BOOTLOADER_SERVICE.getUuid(), OtauClient.BtSmartUuid.OTAU_CSR_OTA_VERSION.getUuid(),
                    mStatemachine.getHandler());
        }
    }

    public void readBootloaderSoftwareVersion() {
        if (mGattClient != null) {
            if (mListener != null) {
                mListener.onOtherLog("readBootloaderSoftwareVersion...");
            }
            mGattClient.requestCharacteristicValue(REQUEST_READ_BL_SOFTWARE_VERSION,
                    OtauClient.BtSmartUuid.OTAU_BOOTLOADER_SERVICE.getUuid(), OtauClient.BtSmartUuid.OTAU_BOOTLOADER_SOFTWARE_VERSION.getUuid(),
                    mStatemachine.getHandler());
        }
    }

    public void readCsrOtaVersion() {
        if (mGattClient != null) {
            mGattClient.requestCharacteristicValue(REQUEST_READ_CSR_OTA_VERSION,
                    OtauClient.BtSmartUuid.OTAU_APPLICATION_SERVICE.getUuid(), OtauClient.BtSmartUuid.OTAU_CSR_OTA_VERSION.getUuid(),
                    mStatemachine.getHandler());
        }
    }

    public void getSoftwareVersion() {
        if (mGattClient != null) {
            if (mListener != null) {
                mListener.onOtherLog("getSoftwareVersion...");
            }
            mGattClient.requestCharacteristicValue(REQUEST_READ_DEVICE_SOFTWARE_VERSION,
                    OtauClient.BtSmartUuid.DEVICE_INFORMATION_SERVICE.getUuid(), OtauClient.BtSmartUuid.SOFTWARE_REVISION.getUuid(),
                    mStatemachine.getHandler());
        }
    }

    /**
     * Read characteristic to get CS keys, e.g BT address ,crystal trim value,Identity Root,Encryption Root
     * 
     * @param requestId
     *            The request id
     * 
     * @param key The index of user key
     */
    public void readCSKeys(int requestId, int key) {
        if (mGattClient != null) {
            if (mListener != null) {
                mListener.onOtherLog("readCSKeys...");
            }
            mGattClient.writeCharacteristicValue(requestId, OtauClient.BtSmartUuid.OTAU_APPLICATION_SERVICE.getUuid(),
                    OtauClient.BtSmartUuid.OTAU_APPLICATION_READCSKEY.getUuid(), new byte[]{(byte) key},
                    mStatemachine.getHandler());
        }
    }

    /**
     * Read CS keys from the Bootloader service, e.g BT address and crystal trim value
     * 
     * @param requestId
     *            The request id
     * @param key
     *            The request key
     */
    public void readCSKeysFromBootloaderService(int requestId, int key) {
        if (mGattClient != null) {
            if (mListener != null) {
                mListener.onOtherLog("readCSKeysFromBootloaderService...");
            }
            mGattClient.writeCharacteristicValue(requestId, OtauClient.BtSmartUuid.OTAU_BOOTLOADER_SERVICE.getUuid(),
                    OtauClient.BtSmartUuid.OTAU_APPLICATION_READCSKEY.getUuid(), new byte[]{(byte) key},
                    mStatemachine.getHandler());
        }
    }

    /**
     * Read characteristic to get CS Block, e.g BT address and crystal trim value
     */
    public void readNextCsBlock() {

        byte data[] = null;

        State.ReadCsBlockState state = State.getReadCsBlockState(mBLEDevice.getAddress());
        Log.d("OtauManager", "readNextCsBlock: start state = " + state);
        if (mListener != null) {
            mListener.onOtherLog("readNextCsBlock: start state = " + state);
        }
        if (state == null) {
            State.setReadCsBlockState(mBLEDevice.getAddress(), State.ReadCsBlockState.READ_CS_BLOCK_BUILD_ID);
            mCskeyList.clear();
            mCskeyList = null;

            if(mIsCancelDownload){
                disconnect();
                mIsCancelDownload = false;
                if (mListener != null) {
                    mListener.onOtherLog("readNextCsBlock: CancelDownload");
                }
                Log.d("OtauManager", "readNextCsBlock: CancelDownload");
                return;
            }

            if(mListener != null)
                mListener.onOtauEnabled(true);
                Log.d("OtauManager", "readNextCsBlock: state == null!!");
            return;
        }

        data = new byte[4];
        if (state == State.ReadCsBlockState.READ_CS_BLOCK_BUILD_ID) {
            data[0] = (byte) ((READ_CSKEY_BUILDID_OFFSET >>> 0) & 0xFF);
            data[1] = (byte) ((READ_CSKEY_BUILDID_OFFSET >>> 8) & 0xFF);
            data[2] = (byte) ((READ_CSKEY_BUILDID_LENGTH >>> 0) & 0xFF);
            data[3] = (byte) ((READ_CSKEY_BUILDID_LENGTH >>> 8) & 0xFF);
        }
        else {
            if (mCskeyList == null) {
                if (mListener != null) {
                    mListener.onOtherLog("readNextCsBlock: mCskeyList is null");
                }
            	Log.d("OtauManager", "readNextCsBlock: mCskeyList is null");
                return;
            }
            
            boolean isFound = false;
            for (CskeyItem cskey : mCskeyList) {
                if (mListener != null) {
                    mListener.onOtherLog("readNextCsBlock: cskey = " + cskey.name + ", id =" + cskey.id);
                }
            	Log.d("OtauManager", "readNextCsBlock: cskey = " + cskey.name + ", id =" + cskey.id);
                if (cskey.id == state.valueOf()) {
                    data[0] = (byte) ((cskey.offset >>> 0) & 0xFF);
                    data[1] = (byte) ((cskey.offset >>> 8) & 0xFF);
                    data[2] = (byte) ((cskey.length * 2 >>> 0) & 0xFF);
                    data[3] = (byte) ((cskey.length * 2 >>> 8) & 0xFF);
                    isFound = true;
                    if (mListener != null) {
                        mListener.onOtherLog("readNextCsBlock: found cs key");
                    }
                    Log.d("OtauManager", "readNextCsBlock: found cs key");
                    break;
                }
            }

            if(!isFound){
                if (mListener != null) {
                    mListener.onOtherLog("readNextCsBlock not Found, try next");
                }
            	Log.d("OtauManager", "readNextCsBlock not Found, try next");
                State.setReadCsBlockState(mBLEDevice.getAddress(), state.getNext());
                readNextCsBlock();
                return;
            }
        }

        if (mGattClient != null) {
            if (mListener != null) {
                mListener.onOtherLog("readNextCsBlock for state " + State.getReadCsBlockState(mBLEDevice.getAddress()));
            }
            Log.d("OtauManager", "readNextCsBlock for state " + State.getReadCsBlockState(mBLEDevice.getAddress()));
            mGattClient.writeCharacteristicValue(REQUEST_READ_CSBLOCK, OtauClient.BtSmartUuid.OTAU_APPLICATION_SERVICE.getUuid(),
                    OtauClient.BtSmartUuid.OTAU_APPLICATION_READCSBLOCK.getUuid(), data, mStatemachine.getHandler());

            State.setOtaState(mBLEDevice.getAddress(), State.OtaState.STATE_OTA_INIT_READ_CS_BLOCK);
        }
        if (mListener != null) {
            mListener.onOtherLog("readNextCsBlock: end");
        }
        Log.d("OtauManager", "readNextCsBlock: end");
    }
    
    public void setNextReadCsBlockState() {
        if (mListener != null) {
            mListener.onOtherLog("setNextReadCsBlockState...");
        }
    	Log.d("OtauManager", "setNextReadCsBlockState");
        State.ReadCsBlockState state = State.getReadCsBlockState(mBLEDevice.getAddress());
        State.setReadCsBlockState(mBLEDevice.getAddress(), state.getNext());
    }

    /**
     * Set current application id for the remote device
     * 
     * @param idx
     *            The index of the application
     */
    public void setApplicationOnChip(int idx) {
        if (mGattClient != null) {
            if (mListener != null) {
                mListener.onOtherLog("setApplicationOnChip...");
            }
            mGattClient.writeCharacteristicValue(REQUEST_SET_APPLCATION_ONCHIP, OtauClient.BtSmartUuid.OTAU_BOOTLOADER_SERVICE.getUuid(),
                    OtauClient.BtSmartUuid.OTAU_CURRENT_APPLICATION.getUuid(), new byte[] { (byte) idx },
                    mStatemachine.getHandler());
        }

    }

    /**
     * Read an authenticated characteristic in application service in order to trigger pairing procedure
     *
     */
    public void readAuthenticatedCharacteristicInApp() {
        if (mGattClient != null) {
            if (mListener != null) {
                mListener.onOtherLog("readAuthenticatedCharacteristicInApp...");
            }
            mGattClient.requestCharacteristicValue(REQUEST_GET_APPLCATION_ONCHIP, OtauClient.BtSmartUuid.OTAU_APPLICATION_SERVICE.getUuid(),
                    OtauClient.BtSmartUuid.OTAU_CURRENT_APPLICATION.getUuid(), mStatemachine.getHandler());
        }

    }

    public boolean isBonded(){
        if(mBLEDevice != null)
            return mBLEDevice.getBondState() == BluetoothDevice.BOND_BONDED;
        else
            return false;
    }

    /**
     * Write characteristic to get BLE device ready for downloading
     */
    public void setTransferControlReady() {
        if (mGattClient != null) {
            mGattClient.writeCharacteristicValue(REQUEST_SET_TRANSFER_CONTROL_READY,
                    OtauClient.BtSmartUuid.OTAU_BOOTLOADER_SERVICE.getUuid(),
                    OtauClient.BtSmartUuid.OTAU_BOOTLOADER_TRANSFER_CONTROL.getUuid(),
                    new byte[] { (byte) SET_TRANSFER_CONTROL_READY }, mStatemachine.getHandler());
        }
    }

    /**
     * Write characteristic to get BLE device into downloading in progress status
     */
    public void setTransferControlInProgress() {
        if (mGattClient != null) {
            if (mListener != null) {
                mListener.onOtherLog("setTransferControlInProgress...");
            }
            mGattClient.writeCharacteristicValue(REQUEST_SET_TRANSFER_CONTROL_INPROGRESS,
                    OtauClient.BtSmartUuid.OTAU_BOOTLOADER_SERVICE.getUuid(),
                    OtauClient.BtSmartUuid.OTAU_BOOTLOADER_TRANSFER_CONTROL.getUuid(),
                    new byte[] { (byte) SET_TRANSFER_CONTROL_INPROGRESS }, mStatemachine.getHandler());
        }
    }

    /**
     * Write characteristic to inform BLE device downloading completed
     */
    private void setTransferComplete() {
        if (mGattClient != null) {
            if (mListener != null) {
                mListener.onOtherLog("setTransferComplete...");
            }
            mGattClient.writeCharacteristicValue(REQUEST_SET_TRANSFER_CONTROL_COMPLETE,
                    OtauClient.BtSmartUuid.OTAU_BOOTLOADER_SERVICE.getUuid(),
                    OtauClient.BtSmartUuid.OTAU_BOOTLOADER_TRANSFER_CONTROL.getUuid(),
                    new byte[] { (byte) SET_TRANSFER_COMPLETED }, mStatemachine.getHandler());

        }
    }

    /**
     * Write characteristic to inform BLE device OTAU failed
     */
    public void setTransferFailed() {
        if (mGattClient != null) {
            mGattClient.writeCharacteristicValue(REQUEST_SET_TRANSFER_CONTROL_FAILED,
                    OtauClient.BtSmartUuid.OTAU_BOOTLOADER_SERVICE.getUuid(),
                    OtauClient.BtSmartUuid.OTAU_BOOTLOADER_TRANSFER_CONTROL.getUuid(),
                    new byte[] { (byte) SET_TRANSFER_FAILED }, mStatemachine.getHandler());
        }
    }

    /**
     * write characteristic to BLE device to pause downloading image action
     */
    public void pauseDownload() {
        if (mGattClient != null) {
            pausedDownload(true);
            byte[] data = new byte[] { (byte) SET_TRANSFER_PAUSED};
            mGattClient.writeCharacteristicValue(REQUEST_PAUSE_DOWNLOAD, OtauClient.BtSmartUuid.OTAU_BOOTLOADER_SERVICE.getUuid(),
                    OtauClient.BtSmartUuid.OTAU_BOOTLOADER_TRANSFER_CONTROL.getUuid(), data, mStatemachine.getHandler());
        }
    }

    /**
     * Write characteristic to BLE device to cancel downloading image action
     */
    public void cancelDownload() {
        if (mGattClient != null) {
            mIsCancelDownload = true;
            byte[] data = new byte[] { (byte) SET_TRANSFER_ABORT };
            mGattClient.writeCharacteristicValue(REQUEST_CANCEL_DOWNLOAD, OtauClient.BtSmartUuid.OTAU_BOOTLOADER_SERVICE.getUuid(),
                    OtauClient.BtSmartUuid.OTAU_BOOTLOADER_TRANSFER_CONTROL.getUuid(), data, mStatemachine.getHandler());
        }
    }

    /**
     * Write characteristic to get current on-chip application id
     */
    public void getApplcationOnChip() {
        if (mGattClient != null) {
            mGattClient.requestCharacteristicValue(REQUEST_GET_APPLCATION_ONCHIP,
                    OtauClient.BtSmartUuid.OTAU_APPLICATION_SERVICE.getUuid(), OtauClient.BtSmartUuid.OTAU_CURRENT_APPLICATION.getUuid(),
                    mStatemachine.getHandler());
        }
    }

    /**
     * Read Data Transfer characteristic
     */
    public void readDataTransferCharacteristic() {
        if (mGattClient != null) {
            if (mListener != null) {
                mListener.onOtherLog("readDataTransferCharacteristic...");
            }
            mGattClient.requestCharacteristicValue(REQUEST_READ_DATA_TRANSFER, OtauClient.BtSmartUuid.OTAU_BOOTLOADER_SERVICE.getUuid(),
                    OtauClient.BtSmartUuid.OTAU_DATA_TRANSFER.getUuid(), mStatemachine.getHandler());
        }
    }

    /**
     * Trigger temporary pairing with the target in bootloader mode
     */
    public void readAuthenticatedCharacteristic() {
        if (mGattClient != null) {
            if (mListener != null) {
                mListener.onOtherLog("readAuthenticatedCharacteristic...");
            }
            mGattClient.requestCharacteristicValue(REQUEST_READ_AUTHENTICATED_CHARACTERISTIC, OtauClient.BtSmartUuid.OTAU_BOOTLOADER_SERVICE.getUuid(),
                    OtauClient.BtSmartUuid.OTAU_DATA_TRANSFER.getUuid(), mStatemachine.getHandler());
        }
    }

    /**
     * Read transfer control status in Bootloader service
     */
    public void readTransferControlStatus() {
        if (mGattClient != null) {
            if (mListener != null) {
                mListener.onOtherLog("readTransferControlStatus...");
            }
            mGattClient.requestCharacteristicValue(REQUEST_READ_TRANSFER_CONTROL_STATUS,
                    OtauClient.BtSmartUuid.OTAU_BOOTLOADER_SERVICE.getUuid(),
                    OtauClient.BtSmartUuid.OTAU_BOOTLOADER_TRANSFER_CONTROL.getUuid(), mStatemachine.getHandler());
        }
    }

    /**
     * Read transfer control status in Bootloader service
     * @param data data
     */
    public void writeDataTransferCharacteristic(byte[] data) {
        if (mGattClient != null) {
            if (mListener != null) {
                mListener.onOtherLog("writeDataTransferCharacteristic...");
            }
            mGattClient.writeCharacteristicValue(REQUEST_WRITE_DATA_TRANSFER, OtauClient.BtSmartUuid.OTAU_BOOTLOADER_SERVICE.getUuid(),
                    OtauClient.BtSmartUuid.OTAU_DATA_TRANSFER.getUuid(), data, mStatemachine.getHandler());
        }
    }

    /**
     * Switch to Bootloader service mode
     */
    public void enableOTAMode() {
        if (mGattClient != null) {
            if (mListener != null) {
                mListener.onOtherLog("enableOTAUMode...");
            }
            mGattClient.writeCharacteristicValue(REQUEST_ENABLE_OTA_MODE, OtauClient.BtSmartUuid.OTAU_APPLICATION_SERVICE.getUuid(),
                    OtauClient.BtSmartUuid.OTAU_CURRENT_APPLICATION.getUuid(), new byte[] { 0 }, mStatemachine.getHandler());
        } else {
            Log.w("OtauManager", "null gatt client");
        }
    }

    /**
     * Register Service Changes indication from BLE device
     * 
     */
    public void registerServiceChangedIndication() {
        if (mGattClient != null) {
            if (mListener != null) {
                mListener.onOtherLog("registerServiceChangedIndication...");
            }
            mGattClient.requestCharacteristicIndication(REQUEST_SERVICE_CHANGED, OtauClient.BtSmartUuid.GATT_SERVICE.getUuid(),
                    OtauClient.BtSmartUuid.SERVICE_CHANGED.getUuid(), mStatemachine.getHandler());
        }
    }

    /**
     * Register Transfer data transfer notification from BLE device
     * @param serviceUuid serviceUuid
     */
    public void registerDataTransferNotification(UUID serviceUuid) {
        if (mGattClient != null) {
            if (mListener != null) {
                mListener.onOtherLog("registerDataTransferNotification...");
            }
            mGattClient.requestCharacteristicNotification(REQUEST_DATATRANSFER_NOTIFICATION, serviceUuid,
                    OtauClient.BtSmartUuid.OTAU_DATA_TRANSFER.getUuid(), mStatemachine.getHandler());
        }
    }

    /**
     * Register Transfer control notification from BLE device
     */
    public void registerBootloaderTransferControlNotifiocation() {
        if (mGattClient != null) {
            if (mListener != null) {
                mListener.onOtherLog("registerBootloaderTransferControlNotifiocation...");
            }
            mGattClient.requestCharacteristicNotification(REQUEST_BL_TRANSFERCONTROL_NOTIFICATION,
                    OtauClient.BtSmartUuid.OTAU_BOOTLOADER_SERVICE.getUuid(),
                    OtauClient.BtSmartUuid.OTAU_BOOTLOADER_TRANSFER_CONTROL.getUuid(), mStatemachine.getHandler());
        }
    }

    /**
     * Merge CS keys into updated image
     * 
     * @param data
     *            byte[] original image data
     * 
     * @param btAddress
     *            String Bluetooth address
     * 
     * @param xlst
     *            int Crystal trim value
     * 
     * @param idRoot
     *            String 32-bit Identity Root value
     * 
     * @param encRoot
     *            String 32-bit Encryption Root value
     * 
     * @return boolean return true if merging is successful
     */
    public boolean setImageData(byte[] data, String btAddress, int xlst, String idRoot, String encRoot) {
        if (data.length <= 0)
            return false;
        if (mListener != null) {
            mListener.onOtherLog("setImageData: btAddress = " + btAddress + ", xlst = " + xlst + ", idRoot = " + idRoot +
                    ", encRoot = " + encRoot);
        }
        Log.d("OtauManager","setImageData: btAddress = " + btAddress + ", xlst = " + xlst + ", idRoot = " + idRoot +
            ", encRoot = " + encRoot);
        try {
            if(mMergedCsKeyList.size() > 0){
                mImageData = mergeKeys(data, mMergedCsKeyList);
                mMergedCsKeyList.clear();
            }else{
                mImageData = mergeKeys(data, btAddress, xlst, idRoot, encRoot);
            }
            
            mProgressing = 0;
        }
        catch (IncorrectImageException e) {
            e.printStackTrace();
            return false;
        }
        catch (UnsupportedEncodingException e) {
            e.printStackTrace();
            return false;
        }
        catch (StringIndexOutOfBoundsException e) {
            e.printStackTrace();
            return false;
        }

        return true;
    }

    /**
     * Host application sends 20-bytes chunk data to BLE device
     * 
     */
    public void sendNextImageChunk() {
        if (mProgressing >= mImageData.length) {
            setTransferComplete();
            mSentNum = 0;
            return;
        }
        else {
            if (!isDownloadpaused) {
                mSentNum = writeBytesToTarget();
            }
        }
    }

    /**
     * Increase progress index if each chunk data is sent out
     * 
     */
    public void increment() {
        if (mSentNum > 0)
            mProgressing += mSentNum;
    }

    /**
     * Reset the number of sending data
     * 
     */
    public void resetSentNum() {
        mSentNum = 0;
    }

    /**
     * Host application pause current downloading
     * @param paused paused
     */
    public void pausedDownload(boolean paused) {
        isDownloadpaused = paused;
    }

    /**
     * Write characteristic to send out 20-byte chunk data
     * 
     */
    private int writeBytesToTarget() {
        Log.d("OtauManager","Writing image data to the target");
        int length = DATA_LENGTH;
        if ((mProgressing + DATA_LENGTH) > mImageData.length) {
            length = mImageData.length - mProgressing;
        }

        final byte[] data = new byte[length];

        System.arraycopy(mImageData, mProgressing, data, 0, length);

        if (mGattClient != null) {
            Log.d("OtauManager","Begin to Write 20 bytes");
			mGattClient.writeCharacteristicValue(REQUEST_WRITE_BYTES_TO_TARGET, OtauClient.BtSmartUuid.OTAU_BOOTLOADER_SERVICE.getUuid(),
                        OtauClient.BtSmartUuid.OTAU_DATA_TRANSFER.getUuid(), data, mStatemachine.getHandler());
            return length;
        }
        else {
            Log.w("OtauManager", "Gatt client object is null");
            return -1;
        }
    }

    /**
     * Update downloading progress
     * 
     */
    void updateProgressbar() {
        if (mListener != null && mImageData.length != 0) {
            mListener.onOtaProgressUpdate(mProgressing * 100 / mImageData.length, mImageData.length, mProgressing);
        }
    }

    /**
     * Generate challenge response key,the mode of operation is Electronic Code Book,
     * and only one block is encrypted at a time.
     * 
     * @param value
     *            byte[] The challenge key from BLE devices
     * 
     * @return byte[] The response key
     */
    byte[] encryptData(byte[] value) {
        try {
            SecretKey aesKey = new SecretKeySpec(mSharedSecretKeySdk, "AES");
            Cipher cipher = Cipher.getInstance("AES/ECB/NoPadding");
            cipher.init(Cipher.ENCRYPT_MODE, aesKey);
            byte[] encrypted = cipher.doFinal(value);
            return encrypted;
        }
        catch (GeneralSecurityException e) {
            e.printStackTrace();
            return null;
        }
    }

    /**
     * Merge CS keys into updated image
     * 
     * @param bb
     *            byte[] original image data
     * 
     * @param bdaddr
     *            String Bluetooth address
     * 
     * @param xtalfreqtrim
     *            int Crystal trim value
     * 
     * @param idRoot
     *            String 32-bit Identity Root value
     * 
     * @param encRoot
     *            String 32-bit EnEncryption Root value
     * 
     * @return boolean Return true if merging is successful
     */
    private byte[] mergeKeys(byte[] bb, String bdaddr, int xtalfreqtrim, String idRoot, String encRoot) throws UnsupportedEncodingException, IncorrectImageException,StringIndexOutOfBoundsException {
    	Log.d("OtauManager","mergeKeys by bootloader");
        return CsConfig.mergeKeys(bb, bdaddr, xtalfreqtrim, idRoot, encRoot);
    }
    
    /**
     * Merge CS keys into updated image
     * 
     * @param bb
     *            byte[] original image data
     * 
     *@param cskeyList
     *            List<CskeyEntry> The list of cskeys which are needed to merge into updated image
     * 
     * @return boolean Return true if merging is successful
     */
    private byte[] mergeKeys(byte[] bb, List<CskeyItem> cskeyList) throws UnsupportedEncodingException, IncorrectImageException,StringIndexOutOfBoundsException {
    	Log.d("OtauManager","mergeKeys by application");
        return CsConfig.mergeKeys(bb, cskeyList);
    }
    
    /**
     * Check remote device if support reading CS Block, this is only in OTAU version 5
     * 
     * @return boolean Return true if remote device supports reading CS Block
     */
    public boolean getCsBlockSupported() {
        return mCsBlockSupported;
    }

    /**
     * Set boolean to indicate remote device support reading CS Block
     * 
     * @param isCsBlockSupported
     *            Boolean which indicate if remote device support reading CS Block
     */
    void setCsBlockSupported(boolean isCsBlockSupported) {
    	Log.d("OtauManager", "setCsBlockSupported: " + isCsBlockSupported);
        this.mCsBlockSupported = isCsBlockSupported;
    }

    /**
     * Get indication if remote device support reading CS Key from Boot-loader Service
     * 
     * @return boolean Return true indicate if remote device support reading CS Keys from Boot-loader Service
     */
    public boolean getReadCskeyFromBootloader() {
        return mReadCskeyFromBootloader;
    }

    /**
     * Set boolean to indicate remote device support reading CS Key from Boot-loader Service
     * 
     * @param isSupported
     *            Boolean which indicate if remote device support reading CS Keys from Boot-loader Service
     */
    void setReadCskeyFromBootloader(boolean isSupported) {
    	Log.d("OtauManager", "setReadCskeyFromBootloader: " + isSupported);
        this.mReadCskeyFromBootloader = isSupported;
    }

    /**
     * Get indication if remote device support software version
     *
     * @return boolean Return true indicate if remote device support software version
     */
    boolean getCsrOtaVersion() {
        return mCsrOtaVersionSupported;
    }

    /**
     * Set boolean to indicate remote device support reading CS Key from Boot-loader Service
     *
     * @param isSupported
     *            Boolean which indicate if remote device support reading CS Keys from Boot-loader Service
     */
    void setCsrOtaVersion(boolean isSupported) {
    	Log.d("OtauManager", "setCsrOtaVersion: " + isSupported);
        this.mCsrOtaVersionSupported = isSupported;
    }

    /**
     * Get indication if remote device support software version
     *
     * @return boolean Return true indicate if remote device support software version
     */
    public boolean getBootloaderSoftwareVersionSupported() {
        return mSoftwareVersionSupported;
    }
    /**
     * Set boolean to indicate remote device support bootloader software version
     *
     * @param exist
     *            Boolean which indicate if remote device support software version
     */
    void setReadBootloaderSoftwareVersion(boolean exist) {
    	Log.d("OtauManager", "setReadBootloaderSoftwareVersion: " + exist);
        this.mSoftwareVersionSupported = exist;
    }

    /**
     * Set CS block offset and reading size
     * 
     * @param list
     *             PSkey entry which is read from uEnergy SDK csk_db.xml file
     */
    public void setCskeyList(ArrayList<CskeyItem> list) {
        mCskeyList = list;
    }
    
    /**
     * Get CS block offset and reading size
     * 
     * @return ArrayList  CSkey item which is read from uEnergy SDK csk_db.xml file
     */
    public ArrayList<CskeyItem>  getCskeyList(){
        return mCskeyList;
    }
    
    /**
     * Add cskey item into merged ArrayList
     * 
     * @param cskey
     *            CskeyItem The CSkey which need to be merged into the image
     */
    public void addMergedCskeyItem(CskeyItem cskey){
        mMergedCsKeyList.add(cskey);
    }

    /**
     * Set a flag to indicate if attributes refresh is done
     *
     * @param done
     *            Boolean which indicate if refreshing attributes is done
     */
    public void setRefreshAttributesDone(boolean done) {
        mRefreshAttributesDone = done;
    }

    /**
     * Get the flag which indicates refreshing attributes is done
     *
     * @return the boolean which indicate refreshing attributes is done
     */
    public boolean getRefreshAttributesDone() {
        return mRefreshAttributesDone;
    }

    /**
     * Refresh local GATT attributes
     *
     */
    public void refreshAttribute() {
        if(mGattClient != null) {
            mGattClient.discoverService(true);
        }
    }

    /**
     * Set the device if it is a mesh device
     *
     * @param meshdevice
     *            Boolean if it is a mesh device
     */
    public void setIsMeshDevice(boolean meshdevice) {
        mIsMeshdevice = meshdevice;
    }

    /**
     * Check if this device is a mesh device
     *
     * @return if this is a mesh device
     */
    public boolean getIsMeshDevice() {
        return mIsMeshdevice;
    }
    
    // csr added start
    // clearClientRequest the GATT operation queue
    public void clearClientRequest() {
    	if (mGattClient != null) {
    		mGattClient.clearRequest();
    	}
    }
    // csr added end
}
