package hk.advanpro.android.sdk.device.ota.dfu;

import hk.advanpro.android.sdk.commons.ALog;
import hk.advanpro.android.sdk.device.Device;
import hk.advanpro.android.sdk.device.ble.BLEBeltDevice;
import hk.advanpro.android.sdk.device.ble.BLEDevice;
import hk.advanpro.android.sdk.device.callback.DeviceCallbackException;
import hk.advanpro.android.sdk.device.callback.DeviceCommandCallback;
import hk.advanpro.android.sdk.device.callback.DeviceDisconnectCallback;
import hk.advanpro.android.sdk.device.ota.BLEFirmwareUpgradeBase;
import hk.advanpro.android.sdk.device.params.ble.belt.BLEBeltIntoBootloaderCommandParams;
import hk.advanpro.android.sdk.device.result.ble.belt.BLEBeltIntoBootloaderCommandResult;

/**
 * DFU 固件升级
 * Created by jalin on 2017/12/15.
 */

public class BLEDfuFirmwareUpgrade extends BLEFirmwareUpgradeBase {
    public static final String TAG = BLEDfuFirmwareUpgrade.class.getSimpleName();
    private String mFilePath;

    @Override
    public void prepareUpgrade() {
        if (getDevice().isApplicationMode()) {
            //发送状态改变事件，进入ota
            onStatusChange(STATUS_INTO_BOOTLOADER);
            //先进入oat模式
            getDevice().command(BLEBeltDevice.CMD_BELT_INTO_BOOTLOADER, new BLEBeltIntoBootloaderCommandParams(), new DeviceCommandCallback<BLEBeltIntoBootloaderCommandResult>() {
                @Override
                public void onSuccess(Device device, BLEBeltIntoBootloaderCommandResult result) {
                    //直接断开连接
                    disconnect();
                }

                @Override
                public void onError(Device device, DeviceCallbackException error) {
                    onUpgradeError(error);
                }
            });
        } else {
            //直接断开连接
            disconnect();
        }
    }

    public void disconnect() {
        //先断开连接
        getDevice().disconnect(new DeviceDisconnectCallback() {
            @Override
            public void onSucceed(Device device) {
                //断开连接，开始升级
                startUpgrade();
            }

            @Override
            public void onError(Device device, DeviceCallbackException error) {
                onUpgradeError(error);
            }
        });
    }

    private void startUpgrade() {
        //开始升级
        mFilePath = getFile().getPath();
        DfuServiceListenerHelper.registerProgressListener(getContext(), mDfuProgressListener);

        final DfuServiceInitiator starter = new DfuServiceInitiator(getAddress())
                .setDeviceName(getDevice().getInfo(BLEDevice.BLE_DEVICE_INFO_NAME))
                .setKeepBond(false)
                .setForceDfu(false)
                .setPacketsReceiptNotificationsEnabled(false)
                .setDisableNotification(true)//禁用通知栏
                .setPacketsReceiptNotificationsValue(DfuServiceInitiator.DEFAULT_PRN_VALUE)
                .setUnsafeExperimentalButtonlessServiceInSecureDfuEnabled(true);
        starter.setZip(mFilePath);
        starter.start(getContext(), DfuUpgradeService.class);

    }


    /**
     * The progress listener receives events from the DFU Service.
     * If is registered in onCreate() and unregistered in onDestroy() so methods here may also be called
     * when the screen is locked or the app went to the background. This is because the UI needs to have the
     * correct information after user comes back to the activity and this information can't be read from the service
     * as it might have been killed already (DFU completed or finished with error).
     */
    private final DfuProgressListener mDfuProgressListener = new DfuProgressListenerAdapter() {
        @Override
        public void onDeviceConnecting(final String deviceAddress) {
            ALog.i(TAG, String.format("[%s] dfu upgrade connecting", deviceAddress));
            onStatusChange(STATUS_RECONNECTED);
        }

        @Override
        public void onDfuProcessStarting(final String deviceAddress) {
            ALog.i(TAG, String.format("[%s] dfu upgrade starting", deviceAddress));
        }

        @Override
        public void onEnablingDfuMode(final String deviceAddress) {
            ALog.i(TAG, String.format("[%s] dfu upgrade switching to dfu", deviceAddress));
        }

        @Override
        public void onFirmwareValidating(final String deviceAddress) {
            ALog.i(TAG, String.format("[%s] dfu upgrade validating", deviceAddress));
        }

        @Override
        public void onDeviceDisconnecting(final String deviceAddress) {
            ALog.i(TAG, String.format("[%s] dfu upgrade disconnecting", deviceAddress));
            onStatusChange(STATUS_DISCONNECTED);
        }

        @Override
        public void onDfuCompleted(final String deviceAddress) {
            ALog.i(TAG, String.format("[%s] dfu upgrade completed", deviceAddress));
            destory();
            //传输完成完成
            onTransmitComplete();
        }

        @Override
        public void onDfuAborted(final String deviceAddress) {
            ALog.i(TAG, String.format("[%s] dfu upgrade aborted", deviceAddress));
        }

        @Override
        public void onProgressChanged(final String deviceAddress, final int percent, final float speed, final float avgSpeed, final int currentPart, final int partsTotal) {
            onTransmitProgress(100, percent);
            ALog.d(TAG, String.format("[%s]dfu uploading progress ,speed =%f,avgSpeed =%f,currentPart=%d，partsTotal=%d,percent=%d", deviceAddress, speed, avgSpeed, currentPart, partsTotal, percent));
        }

        @Override
        public void onError(final String deviceAddress, final int error, final int errorType, final String message) {
            ALog.e(TAG, message);
            destory();
            onUpgradeError(new DeviceCallbackException(error));
        }
    };

    public void destory() {
        DfuServiceListenerHelper.unregisterProgressListener(getContext(), mDfuProgressListener);
    }
}
