package hk.advanpro.android.sdk.device.protocol;


import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.Date;

import hk.advanpro.android.sdk.algorithm.belt.BLEBeltAlg;
import hk.advanpro.android.sdk.device.result.ble.belt.BLEBeltHeartRateEventResult;
import hk.advanpro.android.sdk.device.result.ble.belt.BLEBeltNotifyStepCommandResult;
import hk.advanpro.android.sdk.device.result.ble.belt.BLEBeltNotifyStepEventResult;
import hk.advanpro.android.sdk.device.result.ble.belt.BLEBeltRealSensorEventResult;
import hk.advanpro.android.sdk.device.result.ble.belt.BLEBeltSectionStepCommandResult;
import hk.advanpro.android.sdk.device.result.ble.belt.BLEBeltStepCommandResult;
import hk.advanpro.android.sdk.device.result.ble.belt.BLEBeltStorageClearCommandResult;
import hk.advanpro.android.sdk.device.result.ble.belt.BLEBeltStorageQueryCommandResult;
import hk.advanpro.android.sdk.device.result.ble.belt.BLEBeltStorageStartCommandResult;
import hk.advanpro.android.sdk.device.result.ble.belt.BLEBeltStorageStopCommandResult;
import hk.advanpro.android.sdk.device.result.ble.belt.BleBeltNotifyBreathEventResult;
import hk.advanpro.android.sdk.device.result.ble.belt.BleBeltNotifySleepEventResult;

/**
 * Created by yejianfei on 2017/6/22.
 */


public class BLEBeltProtocol {

    /**
     * @param resister resister
     * @param accX     accX
     * @param accY     accY
     * @param accZ     accZ
     * @param range    织物传感器串接的电阻值(低位在前)，单位为Ω
     * @param accRange 三轴传感器量程
     * @return BLEBeltRealSensorEventResult
     */
    public static BLEBeltRealSensorEventResult buildRealSensorResult(
            byte[] resister, byte[] accX, byte[] accY, byte[] accZ, int range, int accRange) {
        BLEBeltRealSensorEventResult result = new BLEBeltRealSensorEventResult();
        result.setAccRange(accRange);
        result.setRange(range);
        result.setDataTime(new Date());
        result.setSensorRawData(resister);
        result.setAccXRawData(accX);
        result.setAccYRawData(accY);
        result.setAccZRawData(accZ);

        ByteBuffer resisterBuffer = ByteBuffer.wrap(resister).order(ByteOrder.LITTLE_ENDIAN);
        ByteBuffer accXBuffer = ByteBuffer.wrap(accX).order(ByteOrder.LITTLE_ENDIAN);
        ByteBuffer accYBuffer = ByteBuffer.wrap(accY).order(ByteOrder.LITTLE_ENDIAN);
        ByteBuffer accZBuffer = ByteBuffer.wrap(accZ).order(ByteOrder.LITTLE_ENDIAN);

        int i = 0;
        while (resisterBuffer.remaining() > 0) {
            double resisterValue = resisterBuffer.getShort() & 0x0FFFF;
            double accXValue = accXBuffer.getShort();
            double accYValue = accYBuffer.getShort();
            double accZValue = accZBuffer.getShort();

            result.getSensorData()[i] = range * resisterValue / (0xffff - resisterValue);
            result.getAccXData()[i] = accXValue * accRange / (double) 0x8000;
            result.getAccYData()[i] = accYValue * accRange / (double) 0x8000;
            result.getAccZData()[i] = accZValue * accRange / (double) 0x8000;

            i++;
        }

        return result;
    }

    /**
     * 解析心率结果
     *
     * @param data byte[]
     * @return 解析心率结果
     */
    public static BLEBeltHeartRateEventResult buildHeartRateResult(byte[] data) {
        BLEBeltHeartRateEventResult result = new BLEBeltHeartRateEventResult();
        boolean enable = data[0] == (byte) 0x06 && (data[1] & 0xff) != 0xFF;// 0xFF表示正在测算心率
        result.setHeart(enable ? data[1] & 0xff : 0);
        result.setEnable(enable);
        return result;
    }

    public static byte[] buildStepCommand(short year, short month, short day, short hour, short minute, short second, short lastMonth, short lastDay) {
        return new byte[]{
                (byte) 0x02,
                (byte) 0x01,
                (byte) (year > 2000 ? year - 2000 : year),
                (byte) month,
                (byte) day,
                (byte) hour,
                (byte) minute,
                (byte) second,
                (byte) lastMonth,
                (byte) lastDay
        };
    }

    public static BLEBeltStepCommandResult buildStepResult(byte[] data) {
        BLEBeltStepCommandResult result = new BLEBeltStepCommandResult();
        ByteBuffer buffer = ByteBuffer.wrap(data).order(ByteOrder.LITTLE_ENDIAN);
        int length = 1;
        while (buffer.remaining() >= 21 && length > 0) {
            buffer.getShort();
            BLEBeltStepCommandResult.BeltStep step = new BLEBeltStepCommandResult.BeltStep();

            step.setMonth(buffer.get());
            step.setDay(buffer.get());
            step.setWalkSteps(buffer.getInt());
            step.setWalkSeconds(buffer.getInt());
            step.setRunSteps(buffer.getInt());
            step.setRunSeconds(buffer.getInt());
            length = buffer.get() & 0xff;

            result.getSteps().add(step);
        }


        return result;

    }

    public static byte[] buildSectionStepCommand(short year, short month, short day, short begin, short size) {
        ByteBuffer buffer = ByteBuffer.allocate(8);
        buffer.put((byte) 0x02);
        buffer.put((byte) 0x42);
        buffer.put((byte) (year > 2000 ? year - 2000 : year));
        buffer.put((byte) month);
        buffer.put((byte) day);
        buffer.put((byte) begin);
        buffer.put((byte) size);

        return buffer.array();
    }

    public static BLEBeltSectionStepCommandResult buildSectionStepResult(byte[] data) {
        BLEBeltSectionStepCommandResult result = new BLEBeltSectionStepCommandResult();
        ByteBuffer buffer = ByteBuffer.wrap(data).order(ByteOrder.LITTLE_ENDIAN);

        while (buffer.remaining() >= 20) {
            buffer.getShort();

            short walkA = buffer.getShort();
            short runA = buffer.getShort();

            short walkB = buffer.getShort();
            short runB = buffer.getShort();

            short walkC = buffer.getShort();
            short runC = buffer.getShort();

            short walkD = buffer.getShort();
            short runD = buffer.getShort();

            short section = (short) (buffer.getShort() & 0xfff);

            BLEBeltSectionStepCommandResult.BeltStep stepA = new BLEBeltSectionStepCommandResult.BeltStep();
            stepA.setWalkSteps(walkA);
            stepA.setRunSteps(runA);
            stepA.setSection(section);

            result.getSteps().add(stepA);

            BLEBeltSectionStepCommandResult.BeltStep stepB = new BLEBeltSectionStepCommandResult.BeltStep();
            stepB.setWalkSteps(walkB);
            stepB.setRunSteps(runB);
            stepB.setSection(section - 1);

            result.getSteps().add(stepB);

            BLEBeltSectionStepCommandResult.BeltStep stepC = new BLEBeltSectionStepCommandResult.BeltStep();
            stepC.setWalkSteps(walkC);
            stepC.setRunSteps(runC);
            stepC.setSection(section - 2);

            result.getSteps().add(stepC);

            BLEBeltSectionStepCommandResult.BeltStep stepD = new BLEBeltSectionStepCommandResult.BeltStep();
            stepD.setWalkSteps(walkD);
            stepD.setRunSteps(runD);
            stepD.setSection(section - 3);

            result.getSteps().add(stepD);
        }

        return result;
    }

    public static byte[] buildNotifyStepCommand(short year, short month, short day, short hour, short minute, short second, short enable) {
        return new byte[]{
                (byte) 0x02,
                (byte) 0x03,
                (byte) enable,
                (byte) (year > 2000 ? year - 2000 : year),
                (byte) month,
                (byte) day,
                (byte) hour,
                (byte) minute,
                (byte) second
        };
    }

    public static BLEBeltNotifyStepCommandResult buildNotifyStepResult(byte[] data) {
        BLEBeltNotifyStepCommandResult result = new BLEBeltNotifyStepCommandResult();
        result.setEnable(data[2] == 0x01);
        result.setSuccess(data[3] == 0x01);

        return result;
    }

    public static BLEBeltNotifyStepEventResult buildNotifyStepEventResult(byte[] data) {
        BLEBeltNotifyStepEventResult result = new BLEBeltNotifyStepEventResult();
        ByteBuffer buffer = ByteBuffer.wrap(data);
        byte[] bytes = new byte[4];

        buffer.getShort();
        result.setYear((short) (buffer.get() + 2000));
        result.setMonth(buffer.get());
        result.setDay(buffer.get());
        result.setHour(buffer.get());
        result.setMinute(buffer.get());
        result.setSecond(buffer.get());

        buffer.position(8);
        buffer.get(bytes, 0, 3);
        result.setWalkSteps(ByteBuffer.wrap(bytes).order(ByteOrder.LITTLE_ENDIAN).getInt());

        buffer.position(11);
        buffer.get(bytes, 0, 3);
        result.setWalkSeconds(ByteBuffer.wrap(bytes).order(ByteOrder.LITTLE_ENDIAN).getInt());

        buffer.position(14);
        buffer.get(bytes, 0, 3);
        result.setRunSteps(ByteBuffer.wrap(bytes).order(ByteOrder.LITTLE_ENDIAN).getInt());

        buffer.position(17);
        buffer.get(bytes, 0, 3);
        result.setRunSeconds(ByteBuffer.wrap(bytes).order(ByteOrder.LITTLE_ENDIAN).getInt());

        return result;
    }

    public static byte[] buildStorageStartCommand(short year, short month, short day, short hour, short minute, short second, short duration) {
        return new byte[]{
                (byte) 0x05,
                (byte) 0x01,
                (byte) (year > 2000 ? year - 2000 : year),
                (byte) month,
                (byte) day,
                (byte) hour,
                (byte) minute,
                (byte) second,
                (byte) (duration & 0xff),
                (byte) ((duration >> 8) & 0xff),
        };
    }

    public static BLEBeltStorageStartCommandResult buildStorageStartResult(byte[] data) {
        BLEBeltStorageStartCommandResult result = new BLEBeltStorageStartCommandResult();
        ByteBuffer buffer = ByteBuffer.wrap(data).order(ByteOrder.LITTLE_ENDIAN);
        buffer.getShort();

        result.setStatus(buffer.get());
        result.setYear((short) (buffer.get() + 2000));
        result.setMonth(buffer.get());
        result.setDay(buffer.get());
        result.setHour(buffer.get());
        result.setMinute(buffer.get());
        result.setSecond(buffer.get());
        result.setDurationMinute(buffer.getShort());
        result.setTotalMinute(buffer.getShort());
        result.setCode(buffer.get());

        return result;
    }

    public static byte[] buildStorageStopCommand() {
        return new byte[]{
                (byte) 0x05,
                (byte) 0xff
        };
    }

    public static BLEBeltStorageStopCommandResult buildStorageStopResult(byte[] data) {
        BLEBeltStorageStopCommandResult result = new BLEBeltStorageStopCommandResult();
        result.setSuccess(data[2] == 0x01);
        result.setCode(data[3]);

        return result;

    }

    public static byte[] buildStorageQueryCommand() {
        return new byte[]{
                (byte) 0x05,
                (byte) 0x00
        };
    }

    public static byte[] buildIntoBootloaderCommand() {
        return new byte[]{
                (byte) 0x04
        };
    }

    public static BLEBeltStorageQueryCommandResult buildStorageQueryResult(byte[] data) {
        BLEBeltStorageQueryCommandResult result = new BLEBeltStorageQueryCommandResult();
        ByteBuffer buffer = ByteBuffer.wrap(data).order(ByteOrder.LITTLE_ENDIAN);
        buffer.getShort();

        result.setStatus(buffer.get());
        result.setYear((short) (buffer.get() + 2000));
        result.setMonth(buffer.get());
        result.setDay(buffer.get());
        result.setHour(buffer.get());
        result.setMinute(buffer.get());
        result.setSecond(buffer.get());
        result.setDuration(buffer.getShort());
        result.setTotal(buffer.getShort());
        result.setCode(buffer.get());

        return result;
    }

    public static byte[] buildStorageClearCommand() {
        return new byte[]{
                (byte) 0x05,
                (byte) 0x46
        };
    }

    public static BLEBeltStorageClearCommandResult buildStorageClearResult(byte[] data) {
        BLEBeltStorageClearCommandResult result = new BLEBeltStorageClearCommandResult();
        result.setSuccess(data[2] == 0x01);
        result.setCode(data[3]);

        return result;

    }

    public static BleBeltNotifyBreathEventResult buildBreathEventResult(BLEBeltAlg.SmartBelt_TSBreathStats breathStats) {
        BleBeltNotifyBreathEventResult result = new BleBeltNotifyBreathEventResult();
        result.setBreathSN(breathStats.lBreathSN);
        result.setPreTroughIndex(breathStats.lPreTroughIndex);
        result.setPeakIndex(breathStats.lPeakIndex);
        result.setPostTroughIndex(breathStats.lPostTroughIndex);
        result.setInspiratoryTime(breathStats.dInspiratoryTime);
        result.setExspiratoryTime(breathStats.dExspiratoryTime);
        result.setRtoOfIns2Exs(breathStats.dRtoOfIns2Exs);
        result.setBrthTime(breathStats.dBrthTime);
        result.setAmplitude(breathStats.dAmplitude);
        result.setInstantFreq(breathStats.dInstantFreq);
        result.setStartTime(breathStats.SStartTime.getDate());
        result.setEndTime(breathStats.SEndTime.getDate());
        result.setSpanTime(breathStats.SSpanTime.getDate());
        result.setPerDataBuf(breathStats.daPerDataBuf);
        result.setPostureBuf(breathStats.eaPostureBuf);
        result.setAccX(breathStats.daAccX);
        result.setAccY(breathStats.daAccY);
        result.setAccZ(breathStats.daAccZ);
        result.setTotalBrthCnt(breathStats.lTotalBrthCnt);
        result.setTotalBrthTime(breathStats.dTotalBrthTime);
        result.setAvgBrthTime(breathStats.dAvgBrthTime);
        result.setTotalInsTime(breathStats.dTotalInsTime);
        result.setAvgInsTime(breathStats.dAvgInsTime);
        result.setTotalExsTime(breathStats.dTotalExsTime);
        result.setAvgExsTime(breathStats.dAvgExsTime);
        result.setTotalRtoOfIns2Exs(breathStats.dTotalRtoOfIns2Exs);
        result.setAvgRtoOfIns2Exs(breathStats.dAvgRtoOfIns2Exs);
        result.setAvgBrthFreq(breathStats.dAvgBrthFreq);
        result.setTotalBrthAmp(breathStats.dTotalBrthAmp);
        result.setAvgBrthAmp(breathStats.dAvgBrthAmp);


        return result;
    }

    public static BleBeltNotifySleepEventResult buildSleepEventResult(BLEBeltAlg.SleepAnalyzerOut sleepAnalyzerOut) {
        BleBeltNotifySleepEventResult result = new BleBeltNotifySleepEventResult(sleepAnalyzerOut.dSleepGraph, sleepAnalyzerOut.lIntervalOfPoints, sleepAnalyzerOut.iSleepQualityScore);
        return result;
    }


}
